{***************************************************
 Titan Scripter
 Copyright (C) PilotLogic Software House
 http://www.pilotlogic.com

 Package pl_TitanScript.pkg
 This file is part of CodeTyphon Studio
****************************************************}

{$I TitanScript.inc}

unit BASE_Engine;

interface

uses
  LCLIntf, LCLType, LMessages,
{$IFDEF UNIX}
  SyncObjs,
{$ENDIF}
{$IFDEF VARIANTS}
  Variants,
{$ENDIF}
  forms,
  dynlibs,
  SysUtils, Classes, Math, TypInfo,
  BASE_Consts,
  BASE_Sync,
  BASE_Lists,
  BASE_RegExpr;


//====================== CONST ==================================================

const
  _IsDump: Boolean = false;
  MaxTryStack = 100;
  RootNamespaceName = 'NonameNamespace';
  SignLoadOnDemand = true;

  PathDelim  = {$IFDEF WINDOWS} '\'; {$ELSE} '/';{$ENDIF}
  DriveDelim = {$IFDEF WINDOWS} ':'; {$ELSE} '';{$ENDIF}
  PathSep    = {$IFDEF WINDOWS} ';'; {$ELSE} ':';{$ENDIF}

  _SizeVariant = SizeOf(Variant);

  _rmRun = 0;
  _rmStepOver = 1;
  _rmTraceInto = 2;
  _rmRunToCursor = 3;
  _rmTraceToNextSourceLine = 4;
  _ccRegister = 0;
  _ccPascal = 1;
  _ccCDecl = 2;
  _ccStdCall = 3;
  _ccSafeCall = 4;
  _CompiledModuleVersion = 2;

  MaxParams = 20;        // max number of parameters of imported routine
  MaxHash = 97;
  _ssInit = 0;           // component was created, no modules and code
  _ssReadyToCompile = 1; // modules and code were added
  _ssCompiling = 2;      // compiles script
  _ssCompiled = 3;       // all modules were compiled
  _ssLinking = 4;        // links modules in a script
  _ssReadyToRun = 5;     // script was linked and it is ready to run now
  _ssRunning = 6;        // runs script
  _ssPaused = 7;         // script was paused
  _ssTerminated = 8;     // script was terminated

{$IFDEF UNIX}
  varScriptObject = varError;
  varAlias        = $15;
{$ELSE}
  varScriptObject = $0E;
  varAlias        = $15;
{$ENDIF}
  varUndefined = varEmpty;

  FirstSymbolCard = 46;
  DeltaSymbolCard = 64;

  DefaultStackSize = 16000;
  FirstMemSize =  3200;
  DeltaMemSize = 32000;
  FirstProgCard = 64;
  DeltaProgCard = 256;
  FirstStackSize = 128;
  DeltaStackSize = 1024;
//  BOUND_OPER  = -100;
  BOUND_OPER  = 0;
  BOUND_LINES = -30000;
  BOUND_FILES = -50000;
  SP_ROUND_BRACKET_L = -901;
  SP_ROUND_BRACKET_R = -902;
  SP_BRACKET_L = -903;
  SP_BRACKET_R = -904;
  SP_BRACE_L = -905;
  SP_BRACE_R = -906;
  SP_SEMICOLON = -907;
  SP_POINT     = -908;
  SP_COLON     = -909;
  SP_COMMA     = -910;
  SP_EOF       = -911;
  SP_BACKSLASH = -912;

  MAX_VALUE:double=1.7E307;
  MIN_VALUE:double=4.0E-324;
  NEGATIVE_INFINITY:double=-1.7E308;
  POSITIVE_INFINITY:double=1.7E308;
  NaN:double=1.71E308;
  INFINITY:double=1.7E308;

  BR = #13#10;

  typeVOID = 0;
  typeVARIANT = 1;
  typeOLEVARIANT = 2;
  typeBYTE = 3;
  typeCHAR = 4;
  typeBOOLEAN = 5;
  typeWORDBOOL = 6;
  typeLONGBOOL = 7;
  typeINTEGER = 8;
  typeCARDINAL = 9;
  typePOINTER = 10;
  typeDOUBLE = 11;
  typeSTRING = 12;
  typeENUM = 13;
  typeRECORD = 14;
  typeARRAY = 15;
  typeSHORTSTRING = 16;
  typeTEXT = 17;
  typeSUBRANGE = 18;
  typeSET = 19;
  typeCLASS = 20;
  typeCLASSREF = 21;
  typeDYNAMICARRAY = 22;
  typePCHAR = 23;
  typeWORD = 24;
  typeSHORTINT = 25;
  typeSMALLINT = 26;
  typeINT64 = 27;
  typeSINGLE = 28;
  typeCURRENCY = 29;
  typeCOMP = 30;
  typeREAL48 = 31;
  typeEXTENDED = 32;
  typeINTERFACE = 33;
  typeMETHOD = 34;
  typeFILE = 35;
  typeWIDECHAR = 36;
  typeWIDESTRING = 37;
  typePWIDECHAR = 38;
  typeTVarRec = 39;

  IntegerPaxTypes: TIntegerSet = [typeINTEGER, typeINT64, typeCARDINAL,
                                  typeBYTE, typeWORD, typeSMALLINT, typeSHORTINT];

  BooleanPaxTypes: TIntegerSet = [typeBOOLEAN];

  StringPaxTypes: TIntegerSet = [typeSTRING];

  RealPaxTypes: TIntegerSet = [typeDOUBLE, typeSINGLE, typeCURRENCY];

  VarianTTitanTypes: TIntegerSet = [typeVARIANT];

  KindNONE = 0;
  KindVAR = 1;
  KindTYPE = 2;
  KindCONST = 3;
  KindSUB = 4;
  KindREF = 5;
  KindPROP = 6;
  KindLABEL = 7;
  KindHOSTVAR = 8;
  KindHOSTCONST = 9;
  KindHOSTOBJECT = 10;
  KindHOSTINTERFACEVAR = 11;
  KindVIRTUALOBJECT = 12;
  SecsPerHour   = 60 * 60;
  SecsPerDay    = SecsPerHour * 24;
  MSecsPerDay   = SecsPerDay * 1000;
  MSecsPerHour  = SecsPerHour * 1000;

  //............

  SP_INC = -1001;
  SP_DEC = -1002;
  SP_PLUS_ASSIGN = -1003;
  SP_MINUS_ASSIGN = -1004;
  SP_MULT_ASSIGN = -1005;
  SP_DIV_ASSIGN = -1006;
  SP_MOD_ASSIGN = -1007;
  SP_LEFT_SHIFT_ASSIGN = -1008;
  SP_RIGHT_SHIFT_ASSIGN = -1009;
  SP_UNSIGNED_RIGHT_SHIFT_ASSIGN = -1011;
  SP_BITWISE_AND = -1012;
  SP_BITWISE_OR = -1013;
  SP_BITWISE_XOR = -1014;
  SP_LOGICAL_AND = -1015;
  SP_LOGICAL_OR = -1016;
  SP_BITWISE_NOT = -1017;
  SP_LOGICAL_NOT = -1018;
  SP_COND = -1019;
  SP_OR_ASSIGN = -1020;
  SP_AND_ASSIGN = -1021;

type

TIntegerMethodNoParam = function(): Integer of Object;
TIntegerMethodOneParam = function(I: Integer): Integer of Object;

TTitanClassKind = (ckNone, ckClass, ckStructure, ckEnum, ckInterface, ckArray, ckDynamicArray);

TTitanMemberAccess = (maAny, maMyBase, maMyClass);

TTitanModifier = (modDEFAULT, modPUBLIC, modPRIVATE, modSTATIC, modPROTECTED, modVIRTUAL);
TTitanModifierList = set of TTitanModifier;

PVariant = ^Variant;
PInteger = ^TInteger;
PUnknown = ^IUnknown;
PPointer = ^Pointer;

TByteInt = (bb00,bb01,bb02,bb03,bb04,bb05,bb06,bb07,bb08,bb09,
            bb10,bb11,bb12,bb13,bb14,bb15,bb16,bb17,bb18,bb19,
            bb20,bb21,bb22,bb23,bb24,bb25,bb26,bb27,bb28,bb29);

TByteSet = set of TByteInt;


TTitanTypeSub = (tsNone, tsMethod, tsConstructor, tsDestructor, tsGlobal, tsFunction, tsProcedure);

TTitanTokenClass = (tcNone, tcKeyword, tcId, tcSpecial, tcSeparator, tcIntegerConst,
                      tcFloatConst, tcStringConst, tcHtmlStringConst);

TTitanDefKind = (dkClass, dkMethod, dkProperty, dkVariable, dkObject, dkField,
                   dkInterface, dkRecordField, dkConstant, dkRTTIType,
                   dkVirtualObject, dkAny);

TScannerState = (scanText, scanProg);

  TFile = file;
  TTextFile = TextFile;
  TInteger = Integer;
  TReal = Double;
  TSingle  = Single;
  TCurrency = Currency;
  TDouble = Double;
  TReal48 = Double;
  TComp = Double;
  TExtended = Extended;
  TString  = String;
  TShortString = ShortString;
  TByte = Byte;
  TChar = Char;
  TWideChar = WideChar;
  TWideString = WideString;
  TBoolean = Boolean;
  TWordBool = WordBool;
  TLongBool = LongBool;
  TPointer = Pointer;
  TCardinal = Cardinal;
  TWord = Word;
  TShortInt = ShortInt;
  TSmallInt = SmallInt;
  TInt64 = Int64;
  PInt64 = ^TInt64;
  TOleVariant = OleVariant;

  PBoolean = ^TBoolean;
  PDouble = ^TDouble;

TTitanScripterState = Integer;

TCharSet = set of Char;

TTitanToken = record
    Text: String;
    ID: Integer;
    TokenClass: TTitanTokenClass;
    Value: Variant;
    Position: Integer;
  end;

TTitanKinds = TStringList;
TTitanOperators = TStringList;

 TTitanSymbolRec = packed record
    Address: Pointer;
    Position: Integer;
    StartPosition: Integer;
    Next: Integer;
    Level: Integer;
    PName: Integer;
    PType: Integer;
    Count: Integer;
    Local: Integer;
    Kind: Integer;
    Module: Integer;
    CallConv: Integer;
    Misc: Integer;
    Owner: Integer;
    TypeNameIndex: Integer;
    Imported: Boolean;
    Rank: ShortInt;
    Global: Boolean;
    JSIndex: Boolean;
    IsVirtual: Boolean;
  end;


TTitanTryStackRec = record
    B1, B2: Integer;
  end;

TTitanCodeRec = record
    Op, Arg1, Arg2, Res, Vars: Integer;
    PP_Res, PP_Arg1, PP_Arg2: Pointer;
    _Entry, _SubID, SaveOP, AltArg1, LinePos: Integer;
    IsExecutable: Boolean;
  end;

//====================================================


TTitanTypes = class(TStringList)
    function AddType(const TypeName: String; TypeSize: Integer): Integer;
    function GetSize(TypeID: Integer): Integer;
    function GetTypeID(const TypeName: String): Integer;
  end;

TTitanScriptFailure = class(Exception);

TTitanStack = class
  private
    L: Integer;
    function GetItem(I: Integer): Integer;
  public
    Card: Integer;
    fItems: array of Integer;
    constructor Create;
    function  Push(I: Integer): Integer; overload;
    procedure Push(P: Pointer); overload;
    function  Pop: Integer;
    function  PopPointer: Pointer;
    function  Top: Integer;
    procedure SaveToStream(f: TStream);
    procedure LoadFromStream(f: TStream);
    procedure Clear;
    function  StackPtr: Pointer;
    procedure CopyFrom(S: TTitanStack);
    function  IndexOf(I: Integer): Integer;
    function  PushUnique(I: Integer): Integer;
    procedure Delete(I: Integer);
    property  Items[I: Integer]: Integer read GetItem; default;
  end;

TTitanFastStack = class
  private
    function GetItem(I: Integer): Integer;
  public
    Card: Integer;
    fItems: array [0..1024] of Integer;
    constructor Create;
    function  Push(I: Integer): Integer; overload;
    procedure Push(P: Pointer); overload;
    procedure Clear;
    function  Pop: Integer;
    function  PopPointer: Pointer;
    function  Top: Integer;
    function  IndexOf(I: Integer): Integer;
    function  PushUnique(I: Integer): Integer;
    procedure Delete(I: Integer);
    property  Items[I: Integer]: Integer read GetItem; default;
  end;

TTitanCallStack = TTitanStack;
TTitanUsingList = TTitanStack;
TTitanWithStack = TTitanStack;

TTitanIndexedList = class
  private
    fItems: TList;
    function  GetNameID(I: Integer): Integer;
    procedure SetNameID(I: Integer; Value: Integer);
  public
    Objects: TList;
    constructor Create;
    destructor Destroy; override;
    procedure  Clear;
    function  Count: Integer;
    procedure Delete(I: Integer);
    function  AddObject(ID: Integer; AnObject: TObject): Integer;
    function  GetObject(ID: Integer): TObject;
    function  IndexOf(I: Integer): Integer;
    procedure DeleteObject(Index: Integer);
    property  NameID[I: Integer]: Integer read GetNameID write SetNameID; default;
  end;

TTitanHashArray = class
  public
    A: array[0..MaxHash] of TList;
    constructor Create;
    destructor Destroy; override;
    procedure  AddName(const Name: String; NameIndex, _HashNumber: Integer);
    procedure  Clear;
  end;

TTitanHashTable = class
  private
    Keys: array[0..MaxHash] of TList;
    Values: array[0..MaxHash] of TList;
  public
    constructor Create;
    destructor Destroy; override;
    procedure Add(Key: Integer; Value: Integer);
    function  FindValue(Key: Integer; var found: Boolean): Integer;
    procedure DeleteValue(Value: Integer);
    procedure Clear;
  end;

TTitanHashedIndexedList = class(TTitanIndexedList)
    HashTable: TTitanHashTable;
    constructor Create;
    destructor Destroy; override;
    function IndexOf(ID: Integer): Integer;
    function AddObject(ID: Integer; AnObject: TObject): Integer;
    function GetObject(ID: Integer): TObject;
    procedure DeleteObject(Index: Integer);
    procedure Clear;
  end;

TTitanEntryRec = class
  public
    BreakLabel, ContinueLabel: Integer;
    StringLabel: String;
  end;

TTitanEntryStack = class(TList)
    procedure Push(ABreakLabel, AContinueLabel: Integer; var AStringLabel: String);
    procedure Pop;
    function  TopBreakLabel(const AStringLabel: String = ''): Integer;
    function  TopContinueLabel(const AStringLabel: String = ''): Integer;
  end;

TTitanIniFile = class
  private
    L: TStringList;
    FileName: String;
    function  IndexOf(const Key: String): Integer;
    function  GetValue(const Key: String): String;
    procedure SetValue(const Key, Value: String);
  public
    constructor Create(const FileName: String);
    destructor Destroy; override;
    property Values[const Key: String]: String read GetValue write SetValue; default;
  end;

TTitanNameList = class(TStringList)
  public
    HashArray: TTitanHashArray;
    constructor Create;
    destructor Destroy; override;
    function Add(const S: string): Integer; override;
    function GetSize: Integer;
  end;

TTitanLanCodeRec = class
  public
    Key: String;
    LanguageName: String;
    Code: String;
  end;

TTitanCodeList = class(TList)
  private
    function GetRecord(Index: Integer): TTitanLanCodeRec;
  public
    procedure Clear; override;
    function IndexOfKey(const Key: String): Integer;
    function AddCode(const Key, LanguageName, Code: String): TTitanLanCodeRec;
    function GetCode(LanguageName: String): String;
    property Records[Index: Integer]: TTitanLanCodeRec read GetRecord; default;
  end;

TTitanParamList = class
  private
    L: TStringList;
  public
    function  GetParam(const ParamName: String): Variant;
    procedure SetParam(const ParamName: String; const Value: Variant);
    function  GetAddress(const ParamName: String): Pointer;
    function  HasAddress(P: Pointer): Boolean;
    constructor Create;
    destructor Destroy; override;
  end;

TTitanIds = class(TList)
  private
    function GetID(Index: Integer): Integer;
    procedure PutID(Index: Integer; Value: Integer);
  public
    DupYes: Boolean;
    constructor Create(DupYes: Boolean);
    function Add(ID: Integer): Integer;
    function IndexOf(ID: Integer): Integer;
    procedure SaveToStream(S: TStream);
    procedure LoadFromStream(S: TStream);
    property Ids[Index: Integer]: Integer read GetID write PutID; default;
  end;

TTitanVariantStack = class
  private
    L: Integer;
    A: array of Variant;
  public
    Card: Integer;
    constructor Create;
    destructor Destroy; override;
    procedure Clear;
    function Push(const V: Variant): PVariant;
    procedure Pop;
    function Top: Variant;
  end;

TTitanVarList = class
  private
    fItems: TList;
    function GetCount: Integer;
    function Get(Index: Integer): Variant;
  public
    constructor Create;
    destructor Destroy; override;
    function GetAddress(Index: Integer): PVariant;
    function Add(const Value: Variant): Integer;
    function IndexOf(const Value: Variant): Integer;
    procedure Delete(Index: Integer);
    property Count: Integer read GetCount;
    property Items[I: Integer]: Variant read Get; default;
  end;

TTitanAssocRec = class
  public
    Data: TVarRec;
    Obj: TObject;
  end;

TTitanAssocList = class
  private
    fItems: TList;
    function GetItem(I: Integer): TTitanAssocRec;
  public
    constructor Create;
    destructor Destroy; override;
    procedure Clear;
    function Add(const Data: TVarRec; Obj: TObject): Integer;
    function FindObject(const Data: TVarRec): TObject;
    property Items[I: Integer]: TTitanAssocRec read GetItem;
  end;

  TDefaultParameterRec = class
  public
    SubID: Integer;
    ID: Integer;
    Value: Variant;
    procedure SaveToStream(S: TStream);
    procedure LoadFromStream(S: TStream);
  end;

  TDefaultParameterList = class(TList)
  private
    function GetRecord(I: Integer): TDefaultParameterRec;
  public
    destructor Destroy; override;
    procedure AddParameter(SubID, ID: Integer; const Value: Variant);
    function FindFirst(SubID: Integer): Integer;
    function FindNext(I, SubID: Integer): Integer;
    procedure SaveToStream(S: TStream);
    procedure LoadFromStream(S: TStream);
    property Records[I: Integer]: TDefaultParameterRec read GetRecord; default;
  end;

TTitanIDRec = class
  public
    ID, N, Pos: Integer;
    constructor Create(ID, N, Pos: Integer);
  end;

TTitanIDRecList = class(TTitanIndexedList)
  private
    function GetRecord(I: Integer): TTitanIDRec;
  public
    property Records[I: Integer]: TTitanIDRec read GetRecord; default;
  end;

TTitanCallRec = class
  public
    CallN, CallP: Integer;
    ArgsN, ArgsP: TTitanIds;
    constructor Create;
    destructor Destroy; override;
  end;

TTitanCallRecList = class(TTitanIndexedList)
  private
    function GetRecord(I: Integer): TTitanCallRec;
  public
    procedure AddObject(N: Integer; X: TTitanCallRec);
    function Top: TTitanCallRec;
    property Records[I: Integer]: TTitanCallRec read GetRecord; default;
  end;

TTitanAssociativeList = class
  private
    L1, L2: TTitanIds;
    function GetFirst(I: Integer): Integer;
    function GetSecond(I: Integer): Integer;
  public
    constructor Create(DupYes: Boolean);
    destructor Destroy; override;
    function Add(I1, I2: Integer): Integer;
    function Count: Integer;
    procedure Clear;
    function Convert(ID: Integer): Integer;
    property First[I: Integer]: Integer read GetFirst;
    property Second[I: Integer]: Integer read GetSecond;
  end;

  TOperatorList = class
  public
     fItems: TStringList;
     constructor Create;
     destructor Destroy; override;
     procedure Add(const Name: String; Op: Integer);
     function GetName(Op: Integer): String;
     function IndexOf(const Name: String): Integer;
   end;

TTitanObjectList = class
  private
    fItems: TList;
  public
    constructor Create;
    destructor Destroy; override;
    procedure Add(X: TObject);
    procedure Clear;
    procedure Delete(I: Integer);
    function Count: Integer;
    property Items: TList read fItems;
  end;

//===================== Some Varables ================================
var
  OP_SEPARATOR,
  OP_NOP,
  OP_SKIP,
  OP_HALT,
  OP_HALT_GLOBAL,
  OP_HALT_OR_NOP,
  OP_PRINT,
  OP_PRINT_HTML,
  OP_GO,
  OP_GO_FALSE,
  OP_GO_FALSE_EX,
  OP_GO_TRUE,
  OP_GO_TRUE_EX,
  OP_SET_LABEL,
  OP_BEGIN_METHOD,
  OP_CREATE_ARRAY,
  OP_CREATE_DYNAMIC_ARRAY_TYPE,
  OP_GET_FIELD,
  OP_CREATE_SHORT_STRING,
  OP_DO_NOT_DESTROY,
  OP_CREATE_OBJECT,
  OP_CREATE_RESULT,
  OP_CHECK_CLASS,
  OP_SET_TYPE,
  OP_DESTROY_TEMP_OBJECTS,
  OP_DESTROY_HOST,
  OP_DESTROY_OBJECT,
  OP_DESTROY_LOCAL_VAR,
  OP_DESTROY_INTF,
  OP_RELEASE,
  OP_CREATE_REF,
  OP_USE_NAMESPACE,
  OP_END_OF_NAMESPACE,
  OP_BEGIN_WITH,
  OP_END_WITH,
  OP_EVAL_WITH,
  OP_IN,
  OP_IN_SET,
  OP_INSTANCEOF,
  OP_TYPEOF,
  OP_GET_NEXT_PROP,
  OP_SAVE_RESULT,
  OP_GET_ANCESTOR_NAME,
  OP_ON_USES,
  OP_PUSH,
  OP_PUT_PROPERTY,
  OP_CALL,
  OP_CALL_CONSTRUCTOR,
  OP_TYPE_CAST,
  OP_RET,
  OP_EXIT0,
  OP_EXIT,
  OP_RETURN,
  OP_GET_PARAM_COUNT,
  OP_GET_PARAM,
  OP_GET_PUBLISHED_PROPERTY,
  OP_PUT_PUBLISHED_PROPERTY,
  OP_RET_OPERATOR,
  OP_GET_ITEM,
  OP_PUT_ITEM,
  OP_GET_ITEM_EX,
  OP_PUT_ITEM_EX,
  OP_GET_STRING_ELEMENT,
  OP_PUT_STRING_ELEMENT,
  OP_FINALLY,
  OP_CATCH,
  OP_TRY_ON,
  OP_TRY_OFF,
  OP_THROW,
  OP_DISCARD_ERROR,
  OP_EXIT_ON_ERROR,
  OP_ASSIGN,
  OP_ASSIGN_SIMPLE,
  OP_ASSIGN_ADDRESS,
  OP_GET_TERMINAL,
  OP_ASSIGN_RESULT,
  OP_AND,
  OP_OR,
  OP_XOR,
  OP_NOT,
  OP_LEFT_SHIFT, OP_LEFT_SHIFT_EX,
  OP_RIGHT_SHIFT, OP_RIGHT_SHIFT_EX,
  OP_UNSIGNED_RIGHT_SHIFT, OP_UNSIGNED_RIGHT_SHIFT_EX,
  OP_PLUS, OP_PLUS_EX,
  OP_MINUS, OP_MINUS_EX,
  OP_UNARY_PLUS,
  OP_UNARY_MINUS, OP_UNARY_MINUS_EX,
  OP_MULT, OP_MULT_EX,
  OP_DIV, OP_DIV_EX,
  OP_INT_DIV,
  OP_MOD, OP_MOD_EX,
  OP_POWER,
  OP_LT, OP_LT_EX,
  OP_GT, OP_GT_EX,
  OP_GE, OP_GE_EX,
  OP_LE, OP_LE_EX,
  OP_EQ, OP_EQ_EX,
  OP_NE, OP_NE_EX,
  OP_ID, OP_ID_EX,
  OP_NI, OP_NI_EX,
  OP_IS,
  OP_AS,
  OP_TO_INTEGER,
  OP_TO_STRING,
  OP_TO_BOOLEAN,
  OP_DEFINE,
  OP_DECLARE_ON,
  OP_DECLARE_OFF,
  OP_UPCASE_ON,
  OP_UPCASE_OFF,
  OP_OPTIMIZATION_ON,
  OP_OPTIMIZATION_OFF,
  OP_JS_OPERS_ON,
  OP_JS_OPERS_OFF,
  OP_ZERO_BASED_STRINGS_ON,
  OP_ZERO_BASED_STRINGS_OFF,
  OP_VBARRAYS_ON,
  OP_VBARRAYS_OFF,
  OP_USE_LANGUAGE_NAMESPACE,
//............................
  FOP_INC1,
  FOP_INC2,
  FOP_PLUS_INTEGER1,
  FOP_PLUS_INTEGER2,
  FOP_PLUS_DOUBLE1,
  FOP_PLUS_DOUBLE2,
  FOP_PLUS_STRING1,
  FOP_PLUS_STRING2,
  FOP_MINUS_INTEGER1,
  FOP_MINUS_INTEGER2,
  FOP_MINUS_DOUBLE1,
  FOP_MINUS_DOUBLE2,
  FOP_MULT_INTEGER1,
  FOP_MULT_INTEGER2,
  FOP_MULT_DOUBLE1,
  FOP_MULT_DOUBLE2,
  FOP_DIV_INTEGER1,
  FOP_DIV_INTEGER2,
  FOP_DIV_DOUBLE1,
  FOP_DIV_DOUBLE2,
  FOP_MOD1,
  FOP_MOD2,
  FOP_LT_INTEGER1,
  FOP_LT_INTEGER2,
  FOP_LT_DOUBLE1,
  FOP_LT_DOUBLE2,
  FOP_LE_INTEGER1,
  FOP_LE_INTEGER2,
  FOP_LE_DOUBLE1,
  FOP_LE_DOUBLE2,
  FOP_GT_INTEGER1,
  FOP_GT_INTEGER2,
  FOP_GT_DOUBLE1,
  FOP_GT_DOUBLE2,
  FOP_GE_INTEGER1,
  FOP_GE_INTEGER2,
  FOP_GE_DOUBLE1,
  FOP_GE_DOUBLE2,
  FOP_EQ_INTEGER1,
  FOP_EQ_INTEGER2,
  FOP_EQ_DOUBLE1,
  FOP_EQ_DOUBLE2,
  FOP_NE_INTEGER1,
  FOP_NE_INTEGER2,
  FOP_NE_DOUBLE1,
  FOP_NE_DOUBLE2,
  FOP_BITWISE_AND1,
  FOP_BITWISE_AND2,
  FOP_BITWISE_OR1,
  FOP_BITWISE_OR2,
  FOP_BITWISE_XOR1,
  FOP_BITWISE_XOR2,
  FOP_LOGICAL_AND1,
  FOP_LOGICAL_AND2,
  FOP_LOGICAL_OR1,
  FOP_LOGICAL_OR2,
  FOP_LOGICAL_XOR1,
  FOP_LOGICAL_XOR2,
  FOP_BITWISE_NOT1,
  FOP_BITWISE_NOT2,
  FOP_LOGICAL_NOT1,
  FOP_LOGICAL_NOT2,
  FOP_UNARY_MINUS_INTEGER1,
  FOP_UNARY_MINUS_INTEGER2,
  FOP_UNARY_MINUS_DOUBLE1,
  FOP_UNARY_MINUS_DOUBLE2,
  FOP_SHL1,
  FOP_SHL2,
  FOP_SHR1,
  FOP_SHR2,
  FOP_USHR1,
  FOP_USHR2,
  FOP_ASSIGN,
  FOP_GO_FALSE1,
  FOP_GO_FALSE2,
  FOP_GO_TRUE1,
  FOP_GO_TRUE2,
//.............................
  FOP_PUSH,
  FOP_CALL,
  FOP_PLUS1,
  FOP_PLUS2,
  FOP_MINUS1,
  FOP_MINUS2,
  FOP_MULT1,
  FOP_MULT2,
  FOP_DIV1,
  FOP_DIV2,
  FOP_LT1,
  FOP_LT2,
  FOP_LE1,
  FOP_LE2,
  FOP_GT1,
  FOP_GT2,
  FOP_GE1,
  FOP_GE2,
  FOP_EQ1,
  FOP_EQ2,
  FOP_NE1,
  FOP_NE2,
  OP_DECLARE : Integer;

  Undefined: Variant;
  NameIndex_Create: Integer;
  NameIndex_New: Integer;
  OPResultTypes: array[1..40, 1..40] of Integer;
  AssResultTypes: array[1..40, 1..40] of Boolean;
  PAXKinds: TTitanKinds;
  PAXTypes: TTitanTypes;
  PAXOperators: TTitanOperators;
  SupportedPaxTypes: TIntegerSet;
  OverloadableOperators: TOperatorList;
  RelationalOperators: TOperatorList;

//=======================================================

type

TCompareItems = function (P1, P2: Pointer): Integer;

TTitanDllRec = class
  public
    DllName: string;
    Handle: THandle;
  end;

TTitanDllList = class
  private
    fDllList: TList;
    function GetRecord(I: integer): TTitanDllRec;
    function GetCount: integer;
  public
    constructor Create;
    destructor Destroy; override;
    function FindDllByName(const DllName: string): TTitanDllRec;
    function LoadDll(const DllName: string; scripter: Pointer): TTitanDllRec;
    property Records[I: integer]: TTitanDllRec read GetRecord; default;
    property Count: integer read GetCount;
  end;

TTitanEventHandler = class
  public
    Scripter: Pointer;
    SubID: Integer;
    This: Variant;
    ParamCount: Integer;
    Parameters: array of TVarRec;
    ParamTypes: TStringList;
    _EAX, _EDX, _ECX: LongInt;
    _P: Pointer;
    RetSize: Integer;
    DelphiInstance: TObject;
    PropInfo: PPropInfo;
    HostHandler: TMethod;
    OverrideHandlerMode: Integer;
    constructor Create(Scripter: Pointer; pti: PTypeInfo; SubID: Integer; const This: Variant);
    destructor Destroy; override;
    procedure Invoke;
    procedure HandleEvent;
  end;

 TTitanEventHandlerList = class(TList)
  Public
   destructor Destroy; override;
   procedure  ClearHandlers;
   Procedure  DeleteEvent(aEvent:TTitanEventHandler);
 end;


TTitanMethodBody = class;
TTitanDefinition = class;
TTitanMethodImpl = procedure (MethodBody: TTitanMethodBody);
TTitanDefinitionList = class;

TTitanParameter = class
  private
    Scripter: Pointer;
    function  GetAsString: String;
    procedure SetAsString(const Value: String);
    function  GetAsInteger: Integer;
    procedure SetAsBoolean(Value: Boolean);
    function  GetAsBoolean: Boolean;
    procedure SetAsInteger(Value: Integer);
    function  GetAsDouble: Double;
    procedure SetAsDouble(Value: Double);
    function  GetAsCardinal: Cardinal;
    procedure SetAsCardinal(Value: Cardinal);
    function  GetAsPointer: Pointer;
    procedure SetAsPointer(Value: Pointer);
    function  GetAsTObject: TObject;
    procedure SetAsTObject(Value: TObject);
    function  GetAsVariant: Variant;
    procedure SetAsVariant(const Value: Variant);
    procedure Clear;
  public
    VType: Integer;
    PValue: PVariant;
    constructor Create(AScripter: Pointer);
    destructor Destroy; override;
    property AsInteger: Integer read GetAsInteger write SetAsInteger;
    property AsDouble: Double read GetAsDouble write SetAsDouble;
    property AsCardinal: Cardinal read GetAsCardinal write SetAsCardinal;
    property AsPointer: Pointer read GetAsPointer write SetAsPointer;
    property AsString: String read GetAsString write SetAsString;
    property AsBoolean: Boolean read GetAsBoolean write SetAsBoolean;
    property AsTObject: TObject read GetAsTObject write SetAsTObject;
    property AsVariant: Variant read GetAsVariant write SetAsVariant;
  end;

TTitanParameterList = class(TList)
  public
    Scripter: Pointer;
    constructor Create(AScripter: Pointer);
    procedure Clear; override;
  end;

  PObject = ^TObject;

TTitanMethodBody = class
  private
    fResult: TTitanParameter;
    fParameterList: TTitanParameterList;
    fParamCount: Integer;
    fFieldCount: Integer;
    function GetParameter(I: Integer): TTitanParameter;
    procedure SetParameter(I: Integer; Value: TTitanParameter);
    function GetField(I: Integer): TTitanParameter;
    procedure SetField(I: Integer; Value: TTitanParameter);
    function GetSelf: TObject;
    procedure SetSelf(Value: TObject);
    procedure AddParameters(K: Integer);
  public
    PSelf: Pointer;
    Scripter: Pointer;
    Name: String;
    constructor Create(Scripter: Pointer);
    destructor Destroy; override;
    procedure Clear;
    function DefaultValue: Variant;
    function FindNestedClass(const NamespaceName, ClassName: String): Pointer;
    property ParamCount: Integer read fParamCount write fParamCount;
    property FieldCount: Integer read fFieldCount write fFieldCount;
    property Params[I: Integer]: TTitanParameter read GetParameter write SetParameter;
    property Fields[I: Integer]: TTitanParameter read GetField write SetField;
    property Result: TTitanParameter read fResult write fResult;
    property Self: TObject read GetSelf write SetSelf;
  end;

TTitanDefinition = class
  public
    DefValue: Variant;
  protected
    function GetValue: Variant; virtual;
    function PValue: PVariant; virtual;
    procedure SetValue(const AValue: Variant); virtual;
  public
    Name: String;
    DefKind: TTitanDefKind;
    ml: TTitanModifierList;
    Owner: TTitanDefinition;
    DefList: TTitanDefinitionList;
    Index: Integer;
    UserData: Integer;
    Module: Integer;
    ResultType: String;
    constructor Create(DefList: TTitanDefinitionList;
                       const Name: String; Owner: TTitanDefinition; DefKind: TTitanDefKind);
    procedure Dump(var T: TextFile); virtual; abstract;
    procedure AddToScripter(Scripter: Pointer); virtual;
    procedure AddToScripterList;
    function FullName: String;
    property Value: Variant read GetValue write SetValue;
  end;

TTitanMethodDefinition = class(TTitanDefinition)
  public
    NP: Integer;
    Proc: TTitanMethodImpl;
    PClass: TClass;
    DirectProc: Pointer;
    CallConv: Integer;
    Header: String;
    Types: array of Integer;
    ExtraTypes: array of Integer;
    StrTypes: array of String;
    ParamNames: array of String;
    ByRefs: array of Boolean;
    Consts: array of Boolean;
    TypeSub: TTitanTypeSub;
    Sizes: array of Integer;
    Fake: Boolean;
    NewFake: Boolean;
    IsStatic: Boolean;
    ReturnsDynamicArray: Boolean;
    DefParamList: TTitanVarList;
    IsIntf: Boolean;
    MethodIndex: Integer;
    intf_pti: PTypeInfo;
    Guid: TGuid;
    constructor Create(DefList: TTitanDefinitionList; const Name: String;
                       Proc: TTitanMethodImpl; NP: Integer; Owner: TTitanDefinition;
                       IsStatic: Boolean);
    destructor Destroy; override;
    procedure Dump(var T: TextFile); override;
    function  Duplicate: TTitanMethodDefinition;
    procedure AddToScripter(Scripter: Pointer); override;
    function  GetScriptType(ParamNumber: Integer): Integer;
    function  LoadFromDll(Scripter: Pointer; SubID: Integer): Boolean;
  end;

TTitanClassDefinition = class(TTitanDefinition)
  public
    Ancestor: TTitanDefinition;
    PClass: TClass;
    GetPropDef, PutPropDef: TTitanMethodDefinition;
    ClassKind: TTitanClassKind;
    ElType, ElSize: Integer; // dynamic array
    ElTypeName: String;
    IsStaticArray: Boolean;
    ElMin,ElMax:integer;   // Static array
    IsSet: Boolean; // set
    PtiSet: PTypeInfo;
    RecordSize: Integer; // record;
    pti: PTypeInfo;
    guid: TGuid;
    OwnerList: TStringList;
    constructor Create(DefList: TTitanDefinitionList;
                       const Name: String; Owner, Ancestor: TTitanDefinition);
    destructor Destroy; override;
    procedure Dump(var T: TextFile); override;
    procedure AddToScripter(Scripter: Pointer); override;
    procedure CreateOwnerList;
    function  FindClassRec(Scripter: Pointer): Pointer;
  end;

TTitanFieldDefinition = class(TTitanDefinition)
  public
    PClass: TClass;
    FieldType: String;
    Offset: Integer;
    TypeID: Integer;
    constructor Create(DefList: TTitanDefinitionList; PClass: TClass;
                       const FieldName, FieldType: String; Offset: Integer);
    procedure Dump(var T: TextFile); override;
    procedure AddToScripter(Scripter: Pointer); override;
    function  GetFieldValue(SO: pointer): Variant;
    procedure SetFieldValue(SO: pointer; const Value: Variant);
    function  GetFieldAddress(SO: pointer): Pointer;
  end;

TTitanRecordFieldDefinition = class(TTitanDefinition)
  public
    FieldType: String;
    Offset: Integer;
    TypeID: Integer;
    constructor Create(DefList: TTitanDefinitionList;
                       Owner: TTitanClassDefinition;
                       const FieldName, FieldType: String;
                       Offset: Integer);
    function GetFieldAddress(Scripter, Address: pointer): Pointer;
    function GetFieldValue(Scripter, Address: pointer): Variant;
    procedure SetFieldValue(Scripter, Address: pointer; const Value: Variant);
    procedure Dump(var T: TextFile); override;
    procedure AddToScripter(Scripter: Pointer); override;
  end;

TTitanRTTITypeDefinition = class(TTitanDefinition)
  private
    function GetFinalType: Integer;
  public
    pti: PTypeInfo;
    FinalType: Integer;
    constructor Create(DefList: TTitanDefinitionList; pti: PTypeInfo);
    procedure Dump(var T: TextFile); override;
  end;

TTitanPropertyDefinition = class(TTitanDefinition)
  public
    PClass: TClass;
    ReadDef, WriteDef: TTitanDefinition;
    PropDef: String;
    PropType: String;
    NP: Integer;
    constructor Create(DefList: TTitanDefinitionList;
                       const Name: String; Owner: TTitanDefinition;
                       PClass: TClass; ReadDef, WriteDef: TTitanDefinition);
    procedure Dump(var T: TextFile); override;
    procedure AddToScripter(Scripter: Pointer); override;
  end;

TTitanVariableDefinition = class(TTitanDefinition)
  public
    TypeID: Integer;
    Address: Pointer;
    TypeName: String;
    constructor Create(DefList: TTitanDefinitionList;
                       const Name, TypeName: String;
                       Address: Pointer;
                       Owner: TTitanDefinition);
    function  GetValue(Scripter: Pointer): Variant; reintroduce;
    procedure SetValue(Scripter: Pointer; const AValue: Variant); reintroduce;
    function  GetAddress: Pointer;
    procedure Dump(var T: TextFile); override;
    procedure AddToScripter(Scripter: Pointer); override;
  end;

TTitanObjectDefinition = class(TTitanDefinition)
  public
    Instance: TObject;
    constructor Create(DefList: TTitanDefinitionList;
                       const Name: String;
                       Instance: TObject;
                       Owner: TTitanDefinition);
    procedure Dump(var T: TextFile); override;
    procedure AddToScripter(Scripter: Pointer); override;
  end;

TTitanVirtualObjectDefinition = class(TTitanDefinition)
  public
    constructor Create(DefList: TTitanDefinitionList;
                       const Name: String;
                       Owner: TTitanDefinition);
    procedure Dump(var T: TextFile); override;
    procedure AddToScripter(Scripter: Pointer); override;
  end;

TTitanInterfaceVarDefinition = class(TTitanDefinition)
  public
    PIntf: PUnknown;
    guid: TGUID;
    constructor Create(DefList: TTitanDefinitionList;
                       const Name: String;
                       PIntf: PUnknown;
                       const guid: TGUID;
                       Owner: TTitanDefinition);
    procedure Dump(var T: TextFile); override;
    procedure AddToScripter(Scripter: Pointer); override;
  end;

TTitanConstantDefinition = class(TTitanDefinition)
  public
    TypeID: Integer;
    ck: TTitanClassKind;
    constructor Create(DefList: TTitanDefinitionList;
                       const Name: String;
                       const Value: Variant;
                       Owner: TTitanDefinition);
    function GetAddress: Pointer;
    procedure Dump(var T: TextFile); override;
    procedure AddToScripter(Scripter: Pointer); override;
  end;

TTitanDefinitionList = class(TList)
  public
    _LastIndex: Integer;
    ClassIdx: TTitanIds;
    ConstIdx: TTitanIds;
    Global: Boolean;
    MethodIdx: TStringList;
    FieldIdx: TStringList;
    RTTITypeDefIdx: TStringList;
    ConstantIdx: TStringList;
    GUIDIdx: TStringList;
    fListOfClasses:THClassesList;
    fListOfRTTIs:THGenList;
  private
    function GetRecord(Index: Integer): TTitanDefinition;
    function NewIndex: Integer;
  public
    constructor Create(Global: Boolean);
    destructor Destroy; override;
    function Lookup(const Name: String;
                    DefKind: TTitanDefKind;
                    Owner: TTitanDefinition;
                    Scripter: Pointer = nil): TTitanDefinition;
    function LookupAll(const Name: String;
                       DefKind: TTitanDefKind;
                       Owner: TTitanDefinition;
                       Scripter: Pointer = nil): TList;
    function LookupFullName(const FullName: String): TTitanDefinition;
    procedure Clear; override;
    //--------------------------------------------------------------------------------------
    function FindClassDef(PClass: TClass): TTitanClassDefinition;
    function FindInterfaceTypeDef(const Guid: TGuid): TTitanClassDefinition;
    function FindClassDefByPTI(pti: PTypeInfo): TTitanClassDefinition;
    function FindClassDefByName(const Name: String): TTitanClassDefinition;
    function FindRTTITypeDef(pti: PTypeInfo): TTitanRTTITypeDefinition;
    function FindRTTITypeDefByName(const Name: String): TTitanRTTITypeDefinition;
    function FindObjectDef(Instance: TObject; Owner: TTitanDefinition): TTitanObjectDefinition;
    function FindMethod(const ClassName, MethodName: String): TTitanMethodDefinition;
    function FindField(const ClassName, FieldName: String): TTitanFieldDefinition;
    function  FindProperty(const ClassName,PropertyName:String):TTitanPropertyDefinition;
    Procedure FindRecordFields(aOutlist:TStringList;ClassDef:TTitanClassDefinition);
    Function  FindRecordField(ClassDef:TTitanClassDefinition;Const FieldName:string):TTitanRecordFieldDefinition;

    //--------------------------------------------------------------------------------------
    function AddClassSX(PClass:TClass):TTitanClassDefinition;
    function AddClass2SX(const Name:String; Owner,Ancestor:TTitanClassDefinition; RProp,WProp:TTitanMethodImpl):TTitanClassDefinition;

    function AddClassSXFromType(const Name:String; Owner,Ancestor:TTitanClassDefinition):TTitanClassDefinition;
    function AddRoutineSX(const Header:String; Address:Pointer):TTitanMethodDefinition;
    function AddMethodSX(ClassID:integer; const Header:String; Address:Pointer):TTitanMethodDefinition;
    function AddFakeMethodSX(ClassID:integer; const Header:String; Address:Pointer):TTitanMethodDefinition;
    function AddProperty3SXi(ClassID:Integer; GR,GW:TTitanMethodDefinition; const PropDef:String):TTitanPropertyDefinition;
    function AddProperty3SX(ClassID:Integer; GR,GW:TTitanMethodDefinition; const aName,aResType:String):TTitanPropertyDefinition;
    function AddVariableSX(const Name,TypeName:String; Address:Pointer):TTitanVariableDefinition;
    function AddConstantSX(const Name:String; const Value:Variant):TTitanConstantDefinition;
    function AddArrayTypeSX(const TypeName,ElementTypeName:String;Const aMin,aMax:Integer):TTitanClassDefinition;
    function AddDynArrayTypeSX(const TypeName,ElementTypeName:String):TTitanClassDefinition;
    //------------------------------------------------------------------------------------

    function AddField(PClass: TClass; const FieldName, FieldType: String; Offset: Integer): TTitanFieldDefinition;
    function AddRecordField(Owner: TTitanClassDefinition; const FieldName,
                            FieldType: String;
                            Offset: Integer): TTitanRecordFieldDefinition;
    function AddNamespace(const Name: String; Owner: TTitanClassDefinition): TTitanClassDefinition;
    function AddClass1(const Name: String; Owner, Ancestor: TTitanClassDefinition;
                      ReadProp, WriteProp: TTitanMethodImpl): TTitanClassDefinition;
    function AddClass2(PClass: TClass;
                      Owner: TTitanClassDefinition;
                      ReadProp, WriteProp: TTitanMethodImpl;
                      ClassKind: TTitanClassKind = ckClass): TTitanClassDefinition;
    function AddMethod1(const Header: String; Address: Pointer;
                       Owner: TTitanDefinition; Static: Boolean = false): TTitanMethodDefinition;
    function AddMethod2(PClass: TClass; const Header: String; Address: Pointer;
                       Static: Boolean = false): TTitanMethodDefinition;
    function AddMethod3(const Name: String;
                       Proc: TTitanMethodImpl;
                       NP: Integer;
                       Owner: TTitanClassDefinition;
                       Static: Boolean = false): TTitanMethodDefinition;
    function AddMethod4(PClass: TClass;
                       const Name: String;
                       Proc: TTitanMethodImpl;
                       NP: Integer;
                       Static: Boolean = false): TTitanMethodDefinition;
    function AddMethod5(pti: PTypeInfo; const Header: String): TTitanMethodDefinition;
    function AddMethod6(const Guid: TGuid; const Header: String;
                        InitMethodIndex: Integer = -1): TTitanMethodDefinition;
    function AddFakeMethod(PClass: TClass; const Header: String; Address: Pointer;
                       Static: Boolean = false): TTitanMethodDefinition;
    function AddProperty1(PClass: TClass;
                         const Name: String;
                         ProcRead: TTitanMethodImpl;
                         NPRead: Integer;
                         ProcWrite: TTitanMethodImpl;
                         NPWrite: Integer;
                         Default: Boolean = false): TTitanPropertyDefinition; overload;
    function AddProperty2(PClass: TClass; const Name: String;
                         ReadDef, WriteDef: TTitanMethodDefinition;
                         Default: Boolean = false): TTitanPropertyDefinition; overload;
    function AddProperty3(PClass: TClass; const PropDef: String): TTitanPropertyDefinition; overload;
    function AddInterfaceProperty(const guid: TGUID; const PropDef: String): TTitanPropertyDefinition;
    function AddVariable(const Name, TypeName: String;
                         Address: Pointer;
                         Owner: TTitanClassDefinition;
                         Scripter: Pointer = nil): TTitanVariableDefinition;
    function AddObject(const Name: String;
                       Instance: TObject;
                       Owner: TTitanClassDefinition;
                       Scripter: Pointer = nil): TTitanObjectDefinition;
    function AddVirtualObject(const Name: String;
                              Owner: TTitanClassDefinition;
                              Scripter: Pointer = nil): TTitanVirtualObjectDefinition;
    function AddInterfaceVar(const Name: String;
                                            PIntf: PUnknown;
                                            const guid: TGUID;
                                            Owner: TTitanClassDefinition;
                                            Scripter: Pointer = nil): TTitanInterfaceVarDefinition;
    function AddConstant(const Name: String;
                         const Value: Variant;
                         Owner: TTitanClassDefinition;
                         Static: Boolean = true;
                         ck: TTitanClassKind = ckNone;
                         Scripter: Pointer = nil): TTitanConstantDefinition;
    function AddRTTIType(pti: PTypeInfo): TTitanRTTITypeDefinition;
    function AddInterfaceType(const Name: String; Guid: TGuid;
                              const ParentName: String; ParentGuid: TGUID;
                              Owner: TTitanClassDefinition): TTitanClassDefinition;

    function AddEnumTypeByDef(const TypeDef: WideString): TTitanClassDefinition;
    function AddDynamicArrayType(const TypeName, ElementTypeName: String;
                                 Owner: TTitanDefinition):TTitanClassDefinition;
    function AddRecordType(const TypeName: String;
                           Size: Integer;
                           Owner: TTitanDefinition):TTitanClassDefinition;
    procedure Dump(const FileName: String);

    function GetKind(Index: Integer): Integer;
    function GetCount(Index: Integer): Integer;
    function GetName(Index: Integer): String;
    function GetParamTypeName(Index, ParamIndex: Integer): String;
    function GetParamName(Index, ParamIndex: Integer): String;
    function GetTypeName(Index: Integer): String;
    function GetFullName(Index: Integer): String;
    function GetNameIndex(Index: Integer; Scripter: Pointer): Integer;

    function GetAddress(Scripter: Pointer; Index: Integer): Pointer;
    function GetUserData(Scripter: Pointer; Index: Integer): Integer;
    function GetVariant(Scripter: Pointer; Index: Integer): Variant;
    procedure PutVariant(Scripter: Pointer; Index: Integer; const Value: Variant);
    function GetPVariant(Scripter: Pointer; Index: Integer): PVariant;
    function GetRecordByIndex(Index: Integer): TTitanDefinition;
    procedure Unregister(const Name: String;
                         DefKind: TTitanDefKind;
                         Owner: TTitanDefinition;
                         Scripter: Pointer = nil);
    procedure UnregisterAll(DefKind: TTitanDefKind;
                            Scripter: Pointer);
    procedure UnregisterObject(Instance: TObject; Owner: TTitanDefinition);

    function UnregisterNamespace(const Namespacename: String): Boolean;
    property Records[Index: Integer]: TTitanDefinition read GetRecord;
  end;

TTitanFieldRec = class
  public
    ObjectName: String;
    ObjectType: String;
    FieldName: String;
    FieldType: String;
    Address: Pointer;
  end;

TTitanFieldList = class
  private
    fList: TList;
    function GetRecord(Index: Integer): TTitanFieldRec;
  public
    constructor Create;
    destructor Destroy; override;
    function Count: Integer;
    procedure AddRec(const ObjectName: String;
                     ObjectType: String;
                     FieldName: String;
                     FieldType: String;
                     Address: Pointer);
    function FindRecord(const ObjectName: String;
                        ObjectType: String;
                        FieldName: String): TTitanFieldRec;
    procedure Clear;
    property Records[Index: Integer]: TTitanFieldRec read GetRecord;
  end;

TTitanArray = class(TPersistent)
  private
    N: Integer;
    P: Pointer;
    L: Integer;
    fBounds: array of Integer;
    fIndexes: TTitanIds;
    fOwner: Variant;
    fLastIndex: Integer;
    fInputString: String;
    fIndex: Integer;
    _ElMin,_ElMax:integer;
    function AddressOfElement: Pointer;
    function GetLength: Integer;
    procedure PutLength(Value: Integer);
  public
    Scripter: Pointer;
    TypeID: Integer;
    _ElSize: Integer;
    constructor Create(Bounds: array of Integer; typeID: Integer = typeVARIANT);
    destructor Destroy; override;
    procedure _ReDim;
    function  _GetPtr: PVariant;
    function  _GetPtrEx: PVariant;
    function  _Get: Variant;
    function  _GetEx: Variant;
    procedure _Put(const Value: Variant);
    procedure _PutEx(const Value: Variant);
    function  _CheckIndexes: Boolean;
    procedure ReDim(Bounds: array of Integer);
    procedure ClearIndexes;
    function  AddIndex(I: Integer): Integer;
    procedure InsertIndex(I: Integer);
    function  Get(const Indexes: array of Integer): Variant;
    procedure Put(const Indexes: array of Integer; const Value: Variant);
    function  GetPtr(const Indexes: array of Integer): PVariant;
    function  GetEx(const Indexes: array of Integer): Variant;
    procedure PutEx(const Indexes: array of Integer; const Value: Variant);
    Procedure SetMinMax(const aMin,aMax:integer);
    function  CheckIndexes(const Indexes: array of Integer): Boolean;
    function  Duplicate: Variant;
    function  ToString: String;
    function  HighBound(Dim: Integer): Integer;
    function  Typed: Boolean;
    property  Buffer: Pointer read P;
  published
    property Owner: Variant read fOwner write fOwner;
    property DimCount: Integer read L write L;
    property Length: Integer read GetLength write PutLength;
    property Index: Integer read fIndex write fIndex;
    property LastIndex: Integer read fLastIndex write fLastIndex;
    property InputString: String read fInputString write fInputString;
  end;

TTitanScanner = class;

  TDefRec = class
  public
    word: integer;
    What: string;
    Vis: boolean;
  end;

  TDefStack = class
  private
    fItems: TList;
    function GetItem(I: integer): TDefRec;
  public
    constructor Create;
    destructor Destroy; override;
    procedure Clear;
    function Count: integer;
    procedure Push(word: integer; What: string; Vis: boolean);
    procedure Pop;
    property Items[I: integer]: TDefRec read GetItem; default;
  end;

  TScannerRec = class
  public
    LineNumber, PosNumber: integer;
    c: char;
    P: integer;
    Buff: string;
  end;

  TScannerStack = class
  private
    fItems: TList;
    function GetItem(I: integer): TScannerRec;
  public
    constructor Create;
    destructor Destroy; override;
    procedure Clear;
    function Count: integer;
    procedure Push(Scanner: TTitanScanner);
    procedure Pop(Scanner: TTitanScanner);
    property Items[I: integer]: TScannerRec read GetItem; default;
  end;

TTitanScanner = class
  private
    Scripter: Pointer;
    fScannerState: TScannerState;
    Parser: Pointer;
    procedure SetSourceCode(const Value: string);
    function GetSourceCode: string;
    procedure SetScannerState(Value: TScannerState);
  public
    LineNumber, PosNumber: integer;
    Token, BuffToken: TTitanToken;
    c: char;
    P: integer;
    Buff: string;
    VarNameList: TStringList;
    DefStack: TDefStack;
    ScannerStack: TScannerStack;
    LookForward: boolean;
    constructor Create(Parser: Pointer);
    destructor Destroy; override;
    procedure SetScripter(AScripter: Pointer);
    procedure Reset;
    function  EndOfTagExpected: boolean;
    function  GetText(P1, P2: integer): string;
    function  GetNextChar: char;
    function  LA(N: integer): char;
    procedure ReadToken; virtual;
    function  NextToken: TTitanToken;
    function  Next2Token: TTitanToken;
    procedure ScanIdentifier;
    procedure ScanIdentifierEx(ExtraChars: TCharSet);
    procedure ScanChars(CSet: TCharSet);
    procedure ScanString(ch: char);
    procedure ScanHtmlString(const Ch: string);
    procedure ScanFormatString;
    procedure ScanDigits;
    procedure ScanHexDigits;
    procedure ScanWhiteSpace;
    procedure ScanEOF;
    procedure ScanPlus;
    procedure ScanMinus;
    procedure ScanMult;
    procedure ScanDiv;
    procedure ScanMod;
    procedure ScanGT;
    procedure ScanLT;
    procedure ScanEQ;
    procedure ScanLeftRoundBracket;
    procedure ScanRightRoundBracket;
    procedure ScanLeftBracket;
    procedure ScanRightBracket;
    procedure ScanLeftBrace;
    procedure ScanRightBrace;
    procedure ScanColon;
    procedure ScanSemiColon;
    procedure ScanComma;
    procedure ScanPoint;
    procedure ScanBackslash;
    function  GetRegExpr: string;
    procedure ScanCondDir(Start1: char; Start2: TCharSet);
    function IsEOF: boolean;
    procedure IncLineNumber;
    property SourceCode: string read GetSourceCode write SetSourceCode;
    property ScannerState: TScannerState read fScannerState write SetScannerState;
  end;

TTitanScriptObject = class;
TTitanScriptObjectClass = class of TTitanScriptObject;
TTitanMemberRec = class;
TTitanClassRec = class;
TTitanClassList = class;

TTitanProperty = class
  private
    SO: TTitanScriptObject;
    fUpCaseIndex: Integer;
    ExtraDef: TTitanDefinition;
    Scripter: Pointer;
    Val: Variant;
    function GetValue(NParams: Integer): Variant;
    procedure PutValue(NParams: Integer; const AValue: Variant);
    function GetReadID: Integer;
    function GetWriteID: Integer;
  public
    MemberRec: TTitanMemberRec;
    fValue: PVariant;
    constructor Create(SO: TTitanScriptObject; PValue: PVariant; MemberRec: TTitanMemberRec);
    function IsImported: Boolean;
    function IsStatic: Boolean;
    function Definition: TTitanDefinition;
    function PTerminalValue: PVariant;
    function GetAddress(NParams: Integer): PVariant;
    function GetKind: Integer;
    property ReadID: Integer read GetReadID;
    property WriteID: Integer read GetWriteID;
    property Value[NParams: Integer]: Variant read GetValue write PutValue;
    property UpCaseIndex: Integer read fUpCaseIndex;
  end;

TTitanPropertyList = class(TTitanIndexedList)
  private
    Scripter: Pointer;
    function GetProperty(I: Integer): TTitanProperty;
    function GetName(I: Integer): String;
  public
    constructor Create(Scripter: Pointer);
    function FindProperty(PropertyNameIndex: Integer): TTitanProperty;
    function GetDefaultProperty: TTitanProperty;
    property Names[I: Integer]: String read GetName;
    property Properties[I: Integer]: TTitanProperty read GetProperty;
  end;

TTitanScriptObject = class(TPersistent)
  public
    PropertyList: TTitanPropertyList;
    ClassRec: TTitanClassRec;
    HasFordiddenProperties: Boolean;
    function GetClassDef: TTitanClassDefinition;
  public
    Instance: TObject;
    PClass: TClass;
    RefCount: Integer;
    IsClass: Boolean;
    Intf: IUnknown;
    PIntf: PUnknown;
    ExtraPtr: Pointer;
    ExtraPtrSize: Integer;
    ExternalExtraPtr: Boolean;
    ThreadID: Cardinal;
    _ObjCount: Int64;

    constructor Create(ClassRec: TTitanClassRec); virtual;
    destructor Destroy; override;
    function  HasProperty(const PropertyName: String): Boolean; virtual;
    function  Get(PropertyNameIndex: Integer): TTitanProperty;
    function  SafeGet(PropertyNameIndex: Integer): TTitanProperty; virtual;
    function  Put(PropertyNameIndex: Integer; const Value: Variant; NParams: Integer): TTitanProperty;
    function  GetAddress(PropertyNameIndex: Integer; NParams: Integer): PVariant;
    procedure ClearProperty(PropertyNameIndex: Integer);
    function  GetProperty(PropertyNameIndex: Integer; NParams: Integer): Variant; virtual;
    procedure PutProperty(PropertyNameIndex: Integer; const Value: Variant; NParams: Integer); virtual;
    function  GetDefaultProperty(NParams: Integer): Variant;
    procedure PutDefaultProperty(const Value: Variant; NParams: Integer);
    function  CreateProperty(PropertyNameIndex: Integer;
                            PValue: PVariant;
                            MemberRec: TTitanMemberRec): TTitanProperty;
    procedure PutPublishedProperty(const PropertyName: String; const Value: Variant);
    function  GetPublishedProperty(const PropertyName: String; NParams: Integer): Variant;
    function  HasPublishedProperty(const PropertyName: String): Boolean;

    function  Duplicate: TTitanScriptObject;
    function  ToString: String; virtual;
    function  Scripter: Pointer;
    function  IsImported: Boolean;
    function  DefaultValue: Variant; virtual;
    function  GetPropertyName(Index: Integer): String; virtual;
    function  ExtraInstance: TObject; virtual;
    procedure CallAutoDestructor;
    procedure SetDefaultValue(const V: Variant); virtual;
    procedure FreeExtraPtr;
    property  ClassDef: TTitanClassDefinition read GetClassDef;
  end;

TTitanMemberRec = class
  public
    ID, ReadID, WriteID, InitN: Integer;
    ml: TTitanModifierList;
    Kind: Integer;
    Definition: TTitanDefinition;
    UpCaseIndex: Integer;
    fClassRec: TTitanClassRec;
    IsSource: Boolean;
    IsPublished: Boolean;
    NParams: Integer;
    IsImplementationSection: Boolean;
    constructor Create(ID: Integer; ClassRec: TTitanClassRec);
    destructor Destroy; override;
    function Scripter: Pointer;
    procedure SaveToStream(S: TStream);
    procedure LoadFromStream(S: TStream;
                             DS: Integer = 0; DP: Integer = 0);
    function IsStatic: boolean;
    function IsDefault: Boolean;
    function IsImported: Boolean;
    function IsImportedObject: Boolean;
    procedure CheckAccess;
    function GetName: String;
    function GetNameIndex: Integer;
  end;

TTitanMemberList = class(TTitanIndexedList)
  private
    Owner: TTitanClassRec;
    function GetRecord(Index: Integer): TTitanMemberRec;
    function Scripter: Pointer;
  public
    constructor Create(Owner: TTitanClassRec);
    procedure SaveToStream(S: TStream);
    procedure LoadFromStream(S: TStream;
                             DS: Integer = 0; DP: Integer = 0);
    function GetMemberID(const Name: String; UpCase: Boolean = true): Integer;
    function IndexOfMember(const Name: String; UpCase: Boolean = true): Integer;
    function GetMemberRec(const Name: String; UpCase: Boolean = true): TTitanMemberRec;
    function GetMemberRecByID(MemberID: Integer): TTitanMemberRec;
    function UpperCaseIndexOf(UpCaseIndex: Integer): Integer;
    procedure DeleteMember(const Name: String; UpCase: Boolean = true);
    property Records[Index: Integer]: TTitanMemberRec read GetRecord; default;
  end;

TTitanHeapItem = class(TPersistent)
  public
    constructor Create(H: TTitanObjectList);
  end;

TTitanCompileTimeHeap = class(TTitanObjectList)
  public
    procedure ResetCompileStage;
  end;

TTitanClassRec = class
  public
    NameIdx: Integer;
    UpCaseIndex: Integer;
    AncestorClassRec: TTitanClassRec;
    ClassObject: TTitanScriptObject;
    Name: String;
    OwnerName, AncestorName: String;
    Scripter: Pointer;
    ClassID: Integer;
    ModuleID: Integer;
    ml: TTitanModifierList;
    MemberList: TTitanMemberList;
    ck: TTitanClassKind;
    IsStaticArray: Boolean;
    OwnerClassRec: TTitanClassRec;
    fClassDef: TTitanClassDefinition;
    UsingInitList: TTitanIds;
    AutoDestructorID: Integer;
    isSet: Boolean;
    PtiSet: PTypeInfo;
    fHasRunTimeProperties: Boolean;
    constructor Create(AScripter: Pointer; kc: TTitanClassKind);
    destructor Destroy; override;
    procedure ResetCompileStage;
    procedure SaveToStream(S: TStream);
    procedure LoadFromStream(S: TStream; DS: Integer = 0; DP: Integer = 0);
    function  GetClassDef: TTitanClassDefinition;
    function  AddField(ID: Integer; ml: TTitanModifierList): TTitanMemberRec;
    function  AddMethod(ID: Integer; ml: TTitanModifierList): TTitanMemberRec;
    function  AddProperty(ID: Integer; ml: TTitanModifierList): TTitanMemberRec;
    function  AddNestedClass(ID: Integer; ml: TTitanModifierList): TTitanMemberRec;
    function  CreateScriptObject(const ObjectName: String = ''): TTitanScriptObject;
    procedure CreateClassObject;
    function  GetMember(MemberNameIndex: Integer; UpCaseOn: Boolean = true): TTitanMemberRec;
    function  FindMember(MemberNameIndex: Integer; ma: TTitanMemberAccess; UpCaseOn: Boolean = true): TTitanMemberRec;
    function  FindMemberEx(MemberNameIndex: Integer;
                         ma: TTitanMemberAccess;
                         var FoundInBaseClass: Boolean;
                         UpCaseON: Boolean = true): TTitanMemberRec;
    function  IsNestedClass(ClassID: Integer): Boolean;
    function  FindNestedClassID(ClassNameIndex: Integer): Integer;
    function  AddHostMethod(D: TTitanMethodDefinition): TTitanMemberRec;
    function  AddHostProperty(D: TTitanPropertyDefinition): TTitanMemberRec;
    function  AddHostConstant(D: TTitanConstantDefinition): TTitanMemberRec;
    function  AddHostVariable(D: TTitanVariableDefinition): TTitanMemberRec;
    function  AddHostObject(D: TTitanObjectDefinition): TTitanMemberRec;
    function  AddVirtualObject(D: TTitanVirtualObjectDefinition): TTitanMemberRec;
    function  AddHostInterfaceVar(D: TTitanInterfaceVarDefinition): TTitanMemberRec;
    function  AddHostField(D: TTitanFieldDefinition): TTitanMemberRec;
    function  AddHostRecordField(D: TTitanRecordFieldDefinition): TTitanMemberRec;
    function  FindOverloadedSubID(NameIndex: Integer; ma: TTitanMemberAccess; NP: Integer): Integer;
    procedure FindOverloadedSubList(NameIndex: Integer; ma: TTitanMemberAccess; Ids: TTitanIds);
    function  IsImported: Boolean;
    function  InheritsFromClass(AClassID: Integer): Boolean;
    procedure DeleteMember(ID: Integer);
    function  GetConstructorID: Integer;
    function  GetConstructorIDEx: Integer; overload;
    function  GetConstructorIDEx(const Params: array of const): Integer; overload;
    function  GetClassList: TTitanClassList;
    procedure InitStaticFields;
    function  ModuleName: String;
    function  IsStatic: Boolean;
    function  DelphiClass: TClass;
    function  DelphiClassEx: TClass;
    function  HasPublishedProperty(const PropertyName: String): Boolean;
    function  HasPublishedPropertyEx(const PropertyName: String; var AClassName: String): Boolean;
    function  FindBinaryOperatorID(const OperName: String; T1, T2: Integer): Integer;
    function  FindUnaryOperatorID(const OperName: String; T: Integer): Integer;
    function  HasRunTimeProperties: Boolean;
  end;

TTitanClassList = class(TTitanIndexedList)
  private
    Scripter: Pointer;
    SaveCount: Integer;
    function GetRecord(Index: Integer): TTitanClassRec;
    function GetSourceClassList: TList;
  public
    DelegateClassRec,
    ObjectClassRec, BooleanClassRec, StringClassRec, NumberClassRec,
    DateClassRec, FunctionClassRec, ArrayClassRec,
    ActiveXClassRec: TTitanClassRec;
    constructor Create(AScripter: Pointer);
    destructor Destroy; override;
    procedure SaveToStream(S: TStream; I1, I2: Integer);
    procedure LoadFromStream(S: TStream);
    procedure Reset;
    procedure ResetCompileStage;
    procedure InitRunStage;
    procedure ResetRunStage;
    function AddClass(ClassID: Integer;
                      const AClassName, OwnerName, AncestorName: String;
                      ml: TTitanModifierList; ck: TTitanClassKind; UpCase: boolean): TTitanClassRec;
    function FindClass(ClassID: Integer): TTitanClassRec;
    function FindClassByGuid(const guid: TGUID): TTitanClassRec;
    function FindImportedClass(PClass: TClass): TTitanClassRec;
    function ExistsClassByName(const Name: String): boolean;
    function FindClassByName(const Name: String): TTitanClassRec;
    function FindClassByInstance(Instance: TObject): TTitanClassRec;
    function FindNestedClass(OwnerList: TStringList; const Name: String): TTitanClassRec;
    function CreateScriptObject(ClassID: Integer): TTitanScriptObject;
    function FindMember(ClassID, MemberNameIndex: Integer; ma: TTitanMemberAccess): TTitanMemberRec;
    procedure CreateClassObjects(StartRecNo: Integer);
    procedure InitStaticFields(StartRecNo: Integer);
    procedure AddDefinitionList(L: TTitanDefinitionList; N: Integer);
    function CreateBooleanObject(const AValue: Variant): TTitanScriptObject;
    function CreateNumberObject(const AValue: Variant): TTitanScriptObject;
    function CreateStringObject(const AValue: Variant): TTitanScriptObject;
    function CreateDateObject(const AValue: Variant): TTitanScriptObject;
    function CreateFunctionObject(const AValue: Variant): TTitanScriptObject;
    procedure Dump(const FileName: String);
    property Records[Index: Integer]: TTitanClassRec read GetRecord; default;
  end;

TTitanScriptObjectList = class
  private
    fPaxObjects: TList;
    _Scripter: Pointer;
     function GetCount: Integer;
     function GetItem(I: Integer): TTitanScriptObject;
  public
    constructor Create(_Scripter: Pointer);
    destructor Destroy; override;
    procedure Clear;
    procedure Add(SO: TTitanScriptObject);
    procedure Delete(I: Integer);
    function IndexOfDelphiObject(DelphiInstance: TObject): Integer;
    function FindScriptObject(DelphiInstance: TObject): TTitanScriptObject;
    function FindScriptObjectByIntf(PIntf: PUnknown): TTitanScriptObject;
    procedure ResetCompileStage(FreeHost: boolean = false);
    procedure RemoveObject(SO: TTitanScriptObject);
    procedure RemoveStructure(const V: Variant);
    procedure ResetRunStage(FreeHost: Boolean = false);
    procedure RemoveTail(K: Integer; ThreadID: Cardinal = 0);
    function HasObject(SO: TTitanScriptObject): Boolean;
    property PaxObjects: TList read fPaxObjects;
    property Count: Integer read GetCount;
    property Items[I: Integer]: TTitanScriptObject read GetItem; default;
  end;

TTitanDelegate = class(TTitanHeapItem)
  public
    SubID, N: Integer;
    D: TTitanMethodDefinition;
    fName: String;
    constructor Create(Scripter: Pointer; SubID, N: Integer; D: TTitanMethodDefinition);
    destructor Destroy; override;
  published
    property Name: String read fName write fName;
  end;

TTitanError = class(TPersistent)
  private
    fScriptTime: String;
    fDescription: String;
    fModuleName: String;
    fFileName: String;
    fLine: String;
    fLineNumber: Integer;
    fPosition: Integer;
    fTextPosition: Integer;
    fMethodId: Integer;
    fErrClassType: TClass;
  public
    constructor Create;
  published
    property ScriptTime: String read fScriptTime write fScriptTime;
    property Description: String read fDescription write fDescription;
    property Message: String read fDescription write fDescription;
    property ModuleName: String read fModuleName write fModuleName;
    property FileName: String read fFileName write fFileName;
    property Line: String read fLine write fLine;
    property LineNumber: Integer read fLineNumber write fLineNumber;
    property Position: Integer read fPosition write fPosition;
    property TextPosition: Integer read fTextPosition write fTextPosition;
    property MethodId: Integer read fMethodId write fMethodId;
    property ErrClassType: TClass read fErrClassType write ferrClassType;
  end;

ActiveXObject = class
  public
    D: Variant;
    constructor Create(Scripter: Pointer);
  end;

TTitanSymbolTable = class
  private
    Scripter: Pointer;
    Mem: Pointer;
    fCard: Integer;
    StateStack: TTitanStack;
    procedure SetName(I: Integer; const Name: String);
    function GetName(I: Integer): String;
    function GetFullName(I: Integer): String;
    procedure SetNameIndex(I, Value: Integer);
    function GetNameIndex(I: Integer): Integer;
    procedure SetKind(I, AKind: Integer);
    function GetKind(I: Integer): Integer;
    procedure SetCount(I, ACount: Integer);
    function GetCount(I: Integer): Integer;
    procedure SetRank(I, ARank: Integer);
    function GetRank(I: Integer): Integer;
    procedure SetNext(I, Value: Integer);
    function GetNext(I: Integer): Integer;
    procedure SetModule(I, Value: Integer);
    function GetModule(I: Integer): Integer;
    procedure SetPosition(I, Value: Integer);
    function GetStartPosition(I: Integer): Integer;
    procedure SetStartPosition(I, Value: Integer);
    function GetPosition(I: Integer): Integer;
    procedure SetLevel(I: Integer; Value: Integer);
    function GetLevel(I: Integer): Integer;
    procedure SetType(I, AType: Integer);
    function GetType(I: Integer): Integer;
    procedure SetImported(I: Integer; Value: Boolean);
    function GetImported(I: Integer): Boolean;
    procedure SetGlobal(I: Integer; Value: Boolean);
    function GetGlobal(I: Integer): Boolean;
    procedure SetAddr(I: Integer; Address: Pointer);
    function GetAddr(I: Integer): Pointer;
    function GetStrKind(I: Integer): String;
    procedure SetByRef(ParamID: Integer; Value: Integer);
    function GetByRef(ParamID: Integer): Integer;
    function GetStrType(I: Integer): String;
    procedure SetTypeSub(SubID: Integer; Value: TTitanTypeSub);
    function GetTypeSub(SubID: Integer): TTitanTypeSub;
    function GetMemberAccess(RefID: Integer): TTitanMemberAccess;
    procedure SetMemberAccess(RefID: Integer; Value: TTitanMemberAccess);
    procedure SetCallConv(I, Value: Integer);
    function GetCallConv(I: Integer): Integer;
    procedure SetTypeNameIndex(I, Value: Integer);
    function GetTypeNameIndex(I: Integer): Integer;
    procedure SetCard(Value: Integer);
    procedure CheckMem;
    function NameList: TTitanNameList;
    function GetJSIndex(I: Integer): Boolean;
    procedure SetJSIndex(I: Integer; Value: Boolean);
  public
    A: array of TTitanSymbolRec;
    RootNamespaceID: Integer;
    MemBoundVar: Integer;
    MemSize: Integer;
    CreateCard: Integer;
    CreateMemBoundVar: Integer;
    MaxLocalVars: Integer;
    IDfalse: Integer;
    IDtrue: Integer;
    IDundefined: Integer;
    constructor Create(AScripter: Pointer);
    destructor Destroy; override;
    procedure SaveToStream(S: TStream; I1, I2: Integer);
    procedure LoadFromStream(S: TStream; DS: Integer = 0; DP: Integer = 0);
    procedure Reset;
    procedure ResetCompileStage;
    procedure EraseTail(K: Integer);
    procedure InitRunStage;
    procedure ResetRunStage;
    function AllocateVar(I: Integer): Pointer;
    function GetSizeOf(I: Integer): Integer;
    function AppLabel: Integer;
    procedure LinkVariables(SubID: Integer; HasResult: Boolean);
    function GetParamID(SubID: Integer; N: Integer): Integer;
    function GetResultID(SubID: Integer): Integer;
    function GetThisID(SubID: Integer): Integer;
    function GetDllID(SubID: Integer): Integer;
    function GetDllProcID(SubID: Integer): Integer;
    procedure AllocateSub(SubID: Integer);
    function GetSubCount: Integer;
    procedure AllocateSubroutines;
    procedure DeallocateSub(SubID: Integer);
    function IsLocalVar(SubID, ID: Integer): Boolean;
    procedure SetLocal(ID: Integer);
    function IsLocal(ID: Integer): Boolean;
    procedure Erase(I: Integer);
    procedure SaveState;
    procedure RestoreState;
    procedure ReallocateMem(NewSize: Integer);
    function GetStrVal(I: Integer): String;
    function GetAlias(ID: Integer): Variant;
    function GetParamTypeName(SubID, ParamIndex: Integer): String;
    function GetParamName(SubID, ParamIndex: Integer): String;
    function GetTypeName(ID: Integer): String;
    function GetVariant(ID: Integer): Variant;
    procedure ClearVariant(ID: Integer);
    procedure ClearVariantValue(ID: Integer);
    procedure PutVariant(ID: Integer; const Val: Variant);
    function AppVariant(const Val: Variant; HasAddress: Boolean = true): Integer;
    function AppVariantConst(const Val: Variant; Dup: Boolean = false): Integer;
    function GetUserData(ID: Integer): Integer;
    function CodeNumberConst(Val: Variant): Integer;
    function CodeStringConst(const Val: String): Integer;
    function GetMemberRec(RefID: Integer; ma: TTitanMemberAccess): TTitanMemberRec;
    function GetMemberRecEx(RefID: Integer; ma: TTitanMemberAccess;
                            var FoundInBaseClass: Boolean;
                            var SO: TTitanScriptObject): TTitanMemberRec;
    function IsFormalParamID(ID: Integer): Boolean;
    function GetOverloadedSubID(RefID: Integer; ma: TTitanMemberAccess;
                        ParamCount: Integer): Integer;
    procedure SetReference(RefID: Integer; const Base: Variant; ma: TTitanMemberAccess);
    function AppReference(const Base: Variant; ANameIndex: Integer; ma: TTitanMemberAccess): Integer;
    function GetValue(ID: Integer): Variant;
    procedure PutValue(ID: Integer; const Val: Variant);
    function LookUpID(const Name: String; aLevel: Integer; UpCase: Boolean = true): Integer;
    function LookUpSubID(const Name: String; aLevel: Integer; UpCase: Boolean = true): Integer;
    function LookupConstID(const Value: Variant): Integer;
    function IsOutsideMemAddress(A: Pointer): boolean;
    function IsInsideMemAddress(A: Pointer): boolean;
    function IsExternalAddress(A: Pointer): boolean;
    function IsInternalAddress(A: Pointer): boolean;
    procedure IncCard;
    procedure DecCard;
    procedure CheckLength; overload;
    procedure CheckLength(Value: Integer); overload;
    procedure Dump(const FileName: String);
    procedure SetupSubs(StartRecNo: Integer);
    function TypeCast(TypeID: Integer; const Value: Variant): Variant;
    function IsVirtual(SubId: Integer): Boolean;
    function GetAddressEx(ID: Integer): Pointer;
    procedure InitGlobalVars;
    property Name[I: Integer]: String read GetName write SetName;
    property FullName[I: Integer]: String read GetFullName;
    property Kind[I: Integer]: Integer read GetKind write SetKind;
    property Count[I: Integer]: Integer read GetCount write SetCount;
    property Rank[I: Integer]: Integer read GetRank write SetRank;
    property Next[I: Integer]: Integer read GetNext write SetNext;
    property Module[I: Integer]: Integer read GetModule write SetModule;
    property Position[I: Integer]: Integer read GetPosition write SetPosition;
    property StartPosition[I: Integer]: Integer read GetStartPosition write SetStartPosition;
    property Level[I: Integer]: Integer read GetLevel write SetLevel;
    property PType[I: Integer]: Integer read GetType write SetType;
    property Address[I: Integer]: Pointer read GetAddr write SetAddr;
    property ByRef[I: Integer]: Integer read GetByRef write SetByRef;
    property TypeSub[I: Integer]: TTitanTypeSub read GetTypeSub write SetTypeSub;
    property NameIndex[I: Integer]: Integer read GetNameIndex write SetNameIndex;
    property VariantValue[I: Integer]: Variant read GetValue write PutValue;
    property MemberAccess[RefID: Integer]: TTitanMemberAccess read GetMemberAccess write SetMemberAccess;
    property CallConv[SubID: Integer]: Integer read GetCallConv write SetCallConv;
    property TypeNameIndex[ID: Integer]: Integer read GetTypeNameIndex write SetTypeNameIndex;
    property Imported[I: Integer]: Boolean read GetImported write SetImported;
    property Global[I: Integer]: Boolean read GetGlobal write SetGlobal;
    property JSIndex[I: Integer]: Boolean read GetJSIndex write SetJSIndex;
    property Card: Integer read fCard write SetCard;
  end;


TProc = procedure of object;

TTitanDebugInfo = class(TTitanStack)
    procedure PopProc;
  end;

TTitanTryStack = class
  public
    A: array[1..MaxTryStack] of TTitanTryStackRec;
    Card: Integer;
    constructor Create;
    procedure Clear;
    procedure Push(N1, N2: Integer);
    procedure Pop;
    function Legal(N: Integer): boolean;
  end;

TTitanBreakpoint = class
  public
    LineNumber: Integer;
    Condition: String;
    PassCount: Integer;
    CurrPassCount: Integer;
    constructor Create(LineNumber: Integer;
                       const Condition: String;
                       PassCount: Integer);
  end;

 TTitanBreakpointList = class(TTitanIndexedList)
  private
    function GetBreakpoint(I: Integer): TTitanBreakpoint;
  public
    procedure InitCurrPassCounts;
    procedure AddBreakpoint(LineNumber: Integer;
                            const Condition: String;
                            PassCount: Integer);
    function RemoveBreakpoint(LineNumber: Integer): Boolean;
    property Breakpoints[I: Integer]: TTitanBreakpoint read GetBreakpoint; default;
 end;

TTitanCode = class
  private
    Base: Variant;
  public
    Scripter: Pointer;
    SymbolTable: TTitanSymbolTable;
    ClassList: TTitanClassList;
    Stack: TTitanFastStack;
    Prog: array of TTitanCodeRec;
    Card: Integer;
    ArrProc: array[- 350..BOUND_OPER] of TProc;
    N, ErrorN: Integer;
    SubRunCount: Integer;
    UsingList: TTitanUsingList;
    WithStack: TTitanWithStack;
    LevelStack: TTitanFastStack;
    TryStack: TTitanTryStack;
    BreakPointList: TTitanBreakpointList;
    DebugState: Boolean;
    DebugInfo: TTitanDebugInfo;
    StateStack: TTitanStack;
    RefStack: TTitanFastStack;
    SignFOP, fTerminated, DeclareON, SignRETURN, SignInitStage: Boolean;
    ResultValue: Variant;
    _ParamCount: Integer;
    CurrRunMode: Integer;
    _This: Variant;
    InitializationList, FinalizationList: TTitanIds;
    UpcaseOn: Boolean;
    SignVBARRAYS: Boolean;
    SignZERO_BASED_STRINGS: Boolean;
    Undefined2: Variant;
    SignHaltGlobal: Boolean;
    AssignedUndeclaredList: TTitanIds;
    IsAborted: Boolean;
    VV: Variant;
    TempObjects: TList;
    constructor Create(AScripter: Pointer);
    destructor Destroy; override;
    function  GetUpcase(I: Integer): Boolean;
    procedure SetRef(ID: Integer; const V: Variant; ma: TTitanMemberAccess);
    procedure SaveOP;
    function CurrArg1ID: Integer;
    function CurrArg2ID: Integer;
    function CurrResID: Integer;
    function CurrOP: Integer;
    function IsExecutableSourceLine(const ModuleName: String; L: Integer): Boolean;
    procedure SaveToStream(S: TStream; P1, P2: Integer);
    procedure LoadFromStream(S: TStream; DS: Integer = 0; DP: Integer = 0);
    procedure CheckLength;
    procedure Add(InitOp, InitArg1, InitArg2, InitRes: Integer; IsExecutable: Boolean = false);
    procedure GenAt(P: Integer; InitOp, InitArg1, InitArg2, InitRes: Integer; IsExecutable: Boolean = false);
    procedure SetFOP(FOP: Integer; Ex: Boolean = false);
    procedure SaveState;
    procedure RestoreState;
    function  FindCreateObjectStmt(TypeID: Integer): Integer;
    function  GetLanguageNamespaceID: Integer;
    function  GetModuleID(J: Integer): Integer;
    function  ModuleID: Integer;
    function  LineID: Integer;
    function  NextLineID: Integer;
    function  IsExecutableLine(I: Integer): Boolean;
    function  IsBreakpoint(L: Integer): Boolean;
    procedure Dump(const FileName: String);
    procedure Run(RunMode: Integer = _rmRun; DestLine: Integer = 0);
    function  EvalGetFunction(SubID, NParams: Integer): Integer;
    procedure EvalSetProcedure(SubID, NParams: Integer; AValue: Variant);
    procedure CallSub(SubID, ParamCount: Integer; PThis: PVariant; ResultID: Integer; MakeFOP: Boolean = false);
    procedure CallHostSub(D: TTitanMethodDefinition; PThis, PResult: PVariant; ParamCount: Integer);
    function  PopVariant: Variant;
    function  PopAddress: Pointer;
    function  PopInteger: Integer;
    procedure RaiseException;
    procedure Optimization(StartPos, EndPos: Integer);
    procedure LinkGoTo(StartPos, EndPos: Integer);
    procedure RemoveNops;
    procedure ReplaceID(ID1, ID2: Integer);
    procedure InvokeOnChangedVariable;
    function  NextOp(var NewN: Integer): Integer;
    function  CurrMethodID: Integer;
    procedure InitRunStage;
    procedure ResetRunStage;
    procedure ResetCompileStage;
    procedure OperBinaryOperator;
    procedure OperUnaryOperator;
    procedure OperPrint;
    procedure OperHalt;
    procedure OperHaltGlobal;
    procedure OperHaltOrNop;
    procedure OperNop;
    procedure OperSkip;
    procedure OperPutProperty;
    procedure OperCall;
    procedure OperTypeCast;
    procedure OperPush;
    procedure OperRet;
    procedure OperExit0;
    procedure OperExit;
    procedure OperReturn;
    procedure OperSetLabel;
    procedure OperGetParamCount;
    procedure OperGetParam;
    procedure OperGetPublishedProperty;
    procedure OperPutPublishedProperty;
    procedure OperRetOperator;
    procedure OperExitOnError;
    procedure OperDiscardError;
    procedure OperFinally;
    procedure OperCatch;
    procedure OperTryOn;
    procedure OperTryOff;
    procedure OperThrow;
    procedure OperCreateArray;
    procedure OperDoNotDestroy;
    procedure OperGetItem;
    procedure OperPutItem;
    procedure OperGetItemEx;
    procedure OperPutItemEx;
    procedure OperCreateShortString;
    procedure OperGetField;
    procedure OperGetStringElement;
    procedure OperPutStringElement;
    procedure OperCreateObject;
    procedure OperDestroyTempObjects;
    procedure OperDestroyHost;
    procedure OperDestroyObject;
    procedure OperDestroyLocalVar;
    procedure OperDestroyIntf;
    procedure OperRelease;
    procedure OperCreateRef;
    procedure OperUseNamespace;
    procedure OperEndOfNamespace;
    procedure OperBeginWith;
    procedure OperEndWith;
    procedure OperEvalWith;
    procedure OperIn;
    procedure OperInSet;
    procedure OperInstanceOf;
    procedure OperTypeOf;
    procedure OperGetNextProp;
    procedure OperSaveResult;
    procedure OperGetAncestorName;
    procedure OperGo;
    procedure OperGoFalse;
    procedure OperGoFalseEx;
    procedure OperGoTrue;
    procedure OperGoTrueEx;
    procedure OperAssign;
    procedure OperSetOwner;
    procedure OperAssignResult;
    procedure OperAssignAddress;
    procedure OperGetTerminal;
    procedure OperAnd;
    procedure OperOr;
    procedure OperXor;
    procedure OperNot;
    procedure OperLeftShift;
    procedure OperLeftShift_Ex;
    procedure OperRightShift;
    procedure OperRightShift_Ex;
    procedure OperUnsignedRightShift;
    procedure OperUnsignedRightShift_Ex;
    procedure OperPlus;
    procedure OperPlus_Ex;
    procedure OperMinus;
    procedure OperMinus_Ex;
    procedure OperUnaryPlus;
    procedure OperUnaryMinus;
    procedure OperUnaryMinusEx;
    procedure OperMult;
    procedure OperMult_Ex;
    procedure OperDiv;
    procedure OperDiv_Ex;
    procedure OperIntDiv;
    procedure OperMod;
    procedure OperMod_Ex;
    procedure OperPower;
    procedure OperLT;
    procedure OperLT_Ex;
    procedure OperGT;
    procedure OperGT_Ex;
    procedure OperLE;
    procedure OperLE_Ex;
    procedure OperGE;
    procedure OperGE_Ex;
    procedure OperEQ;
    procedure OperEQ_Ex;
    procedure OperNE;
    procedure OperNE_Ex;
    procedure OperID;
    procedure OperID_Ex;
    procedure OperNI;
    procedure OperNI_Ex;
    procedure OperIS;
    procedure OperAS;
    procedure OperToInteger;
    procedure OperToString;
    procedure OperToBoolean;
    procedure OperDefine;
    procedure OperDeclareOn;
    procedure OperDeclareOff;
    procedure OperUpcaseOn;
    procedure OperUpcaseOff;
    procedure OperOptimizationOn;
    procedure OperOptimizationOff;
    procedure OperVBArraysOff;
    procedure OperVBArraysOn;
    procedure OperZeroBasedStringsOn;
    procedure OperZeroBasedStringsOff;
    procedure FindUnaryOperator(const Name: String; const V2: Variant);
    procedure FindBinaryOperator(const Name: String; const V1, V2: Variant);
//------------------------------------------
    procedure FOperGoFalse1;
    procedure FOperGoFalse2;
    procedure FOperGoTrue1;
    procedure FOperGoTrue2;
    procedure FOperAssign;
    procedure FOperINC1;
    procedure FOperINC2;
    procedure FOperPlusInteger1;
    procedure FOperPlusInteger2;
    procedure FOperPlusDouble1;
    procedure FOperPlusDouble2;
    procedure FOperPlusString1;
    procedure FOperPlusString2;
    procedure FOperMinusInteger1;
    procedure FOperMinusInteger2;
    procedure FOperMinusDouble1;
    procedure FOperMinusDouble2;
    procedure FOperMultInteger1;
    procedure FOperMultInteger2;
    procedure FOperMultDouble1;
    procedure FOperMultDouble2;
    procedure FOperDivInteger1;
    procedure FOperDivInteger2;
    procedure FOperDivDouble1;
    procedure FOperDivDouble2;
    procedure FOperMod1;
    procedure FOperMod2;
    procedure FOperLTInteger1;
    procedure FOperLTInteger2;
    procedure FOperLTDouble1;
    procedure FOperLTDouble2;
    procedure FOperLEInteger1;
    procedure FOperLEInteger2;
    procedure FOperLEDouble1;
    procedure FOperLEDouble2;
    procedure FOperGTInteger1;
    procedure FOperGTInteger2;
    procedure FOperGTDouble1;
    procedure FOperGTDouble2;
    procedure FOperGEInteger1;
    procedure FOperGEInteger2;
    procedure FOperGEDouble1;
    procedure FOperGEDouble2;
    procedure FOperEQInteger1;
    procedure FOperEQInteger2;
    procedure FOperEQDouble1;
    procedure FOperEQDouble2;
    procedure FOperNEInteger1;
    procedure FOperNEInteger2;
    procedure FOperNEDouble1;
    procedure FOperNEDouble2;
    procedure FOperBitwiseAND1;
    procedure FOperBitwiseAND2;
    procedure FOperBitwiseOR1;
    procedure FOperBitwiseOR2;
    procedure FOperBitwiseXOR1;
    procedure FOperBitwiseXOR2;
    procedure FOperLogicalAND1;
    procedure FOperLogicalAND2;
    procedure FOperLogicalOR1;
    procedure FOperLogicalOR2;
    procedure FOperLogicalXOR1;
    procedure FOperLogicalXOR2;
    procedure FOperBitwiseNOT1;
    procedure FOperBitwiseNOT2;
    procedure FOperLogicalNOT1;
    procedure FOperLogicalNOT2;
    procedure FOperUnaryMinusInteger1;
    procedure FOperUnaryMinusInteger2;
    procedure FOperUnaryMinusDouble1;
    procedure FOperUnaryMinusDouble2;
    procedure FOperSHL1;
    procedure FOperSHL2;
    procedure FOperSHR1;
    procedure FOperSHR2;
    procedure FOperUSHR1;
    procedure FOperUSHR2;
//.....................................
    procedure FOperPush;
    procedure FOperPlus1;
    procedure FOperPlus2;
    procedure FOperMinus1;
    procedure FOperMinus2;
    procedure FOperMult1;
    procedure FOperMult2;
    procedure FOperDiv1;
    procedure FOperDiv2;
    procedure FOperGT1;
    procedure FOperGT2;
    procedure FOperGE1;
    procedure FOperGE2;
    procedure FOperLT1;
    procedure FOperLT2;
    procedure FOperLE1;
    procedure FOperLE2;
    procedure FOperEQ1;
    procedure FOperEQ2;
    procedure FOperNE1;
    procedure FOperNE2;
    procedure FOperCall;
    property Terminated: Boolean read fTerminated write fTerminated;
  end;


type
TTitanParser = class;
TTitanParserClass = class of TTitanParser;

TTitanLevelStack = class(TTitanStack)
  private
    Parser: TTitanParser;
    ClassList: TTitanClassList;
    SymbolTable: TTitanSymbolTable;
    StateStack: TTitanStack;
  public
    constructor Create(Parser: TTitanParser);
    destructor Destroy; override;
    procedure SetScripter(AScripter: Pointer);
    function PushSub(SubID, ClassID: Integer; ml: TTitanModifierList): TTitanMemberRec;
    procedure PushClass(ClassID, AncestorClassID: Integer; ml: TTitanModifierList;
                        ck: TTitanClassKind; UpCase: Boolean);
    function PushTempClass: Integer;
    function KindTop: Integer;
    procedure Save;
    procedure Restore;
  end;

  TKeywordList = class(TStringList)
    function IndexOf(const S: string): Integer; override;
  end;

TTitanBaseLanguage = class(TComponent)
  protected
    function GetLongStrLiterals: Boolean; virtual; abstract;
  end;

TTitanParser = class
  private
    function  GetName(ID: Integer): String;
    procedure SetName(ID: Integer; const Value: String);
    function  GetKind(ID: Integer): Integer;
    procedure SetKind(ID: Integer; Value: Integer);
    function  GetNameIndex(ID: Integer): Integer;
    procedure SetNameIndex(ID: Integer; Value: Integer);
    function  GetVariant(ID: Integer): Variant;
    procedure SetVariant(ID: Integer; const Value: Variant);
    function  GetAddress(ID: Integer): Pointer;
    procedure SetAddress(ID: Integer; Value: Pointer);
    procedure SetTypeID(ID: Integer; Value: Integer);
    function  GetTypeID(ID: Integer): Integer;
    procedure SetCount(ID: Integer; Value: Integer);
    function  GetCount(ID: Integer): Integer;
    procedure SetNext(ID: Integer; Value: Integer);
    function  GetNext(ID: Integer): Integer;
    procedure SetTypeSub(SubID: Integer; Value: TTitanTypeSub);
    function  GetTypeSub(SubID: Integer): TTitanTypeSub;
  public
    Scripter: Pointer;
    ModuleID: Integer;
    SymbolTable: TTitanSymbolTable;
    Scanner: TTitanScanner;
    ClassList: TTitanClassList;
    Code: TTitanCode;
    CurrToken: TTitanToken;
    UpCase: boolean;
    Keywords: TKeywordList;
    LevelStack: TTitanLevelStack;
    WithStack: TTitanWithStack;
    EntryStack: TTitanEntryStack;
    UsingList: TTitanUsingList;
    TempObjectList: TTitanIds;
    ArrayArgumentList: TTitanIds;
    LocalVars: TTitanIds;
    StatementLabel: String;
    TempCount: Integer;
    DeclareSwitch: boolean;
    FieldSwitch: boolean;
    DirectiveSwitch: boolean;
    OptionExplicit: Boolean;
    ShortEvalSwitch: Boolean;
    NewID: boolean;
    ArgumentListSwitch: Boolean;
    LanguageName, FileExt, IncludeFileExt: String;
    _Language: TTitanBaseLanguage;
    DefaultCallConv: Integer;
    NamespaceAsModule: Boolean;
    SyntaxCheckOnly: Boolean;
    JavaScriptOperators: Boolean;
    DeclareVariables: Boolean;
    IsArrayInitialization: Boolean;
    VBArrays: Boolean;
    ZeroBasedStrings: Boolean;
    Backslash: Boolean;
    BlockCount: Integer;
    IsImplementationSection: Boolean;
    IsExecutable: Boolean;
    DuplicateVars: Boolean;
    WithCount: Integer;
    constructor Create; virtual;
    destructor Destroy; override;
    procedure BeginBlock;
    procedure EndBlock;
    procedure SetScripter(AScripter: Pointer);
    procedure Reset; virtual;
    procedure AddExtraCode(const Key, StrCode: String);
    function  NewLabel: Integer;
    function  NewField(const FieldName: String): Integer;
    function  NewVar: Integer; overload;
    function  NewVar(const V: Variant): Integer; overload;
    function  NewConst(const Value: Variant): Integer;
    function  NewRef: Integer;
    function  IsCurrText(const S: String): boolean;
    function  IsNextText(const S: String): boolean;
    function  IsNext2Text(const S: String): boolean;
    function  NextToken: TTitanToken;
    function  Next2Text: String;
    procedure Match(const S: String); virtual;
    procedure Call_SCANNER; virtual;
    function  Gen(Op, Arg1, Arg2, Res: Integer): Integer; virtual;
    procedure GenAt(N: Integer; Op, Arg1, Arg2, Res: Integer);
    procedure LinkVariables(SubID: Integer; HasResult: Boolean);
    procedure GenRef(Arg1: Integer; ma: TTitanMemberAccess; Res: Integer);
    function  GenEvalWith(ID: Integer): Integer;
    function InUsing(const MemberName: String): Integer;
    function  GenBeginWith(ID: Integer): Integer;
    procedure GenEndWith(WithCount: Integer);
    procedure GenHtml;
    procedure SetLabelHere(L: Integer);
    function IsConstant: boolean; virtual;
    function IsKeyword(const S: String): boolean;
    function CurrLevel: Integer;
    function IsNestedSub(SubID: Integer): Boolean;
    function GetOuterSubID(SubID: Integer): Integer;
    function LookUpID(const Name: String): Integer;
    function LookUpLocalID(const Name: String): Integer;
    function IsLabelId: boolean;
    function LA(I: Integer): Char;
    function CurrClassID: Integer;
    function CurrMethodID: Integer;
    function CurrThisID: Integer;
    function CurrResultID: Integer;
    function CurrSubID: Integer;
    function CurrClassRec: TTitanClassRec;
    function ToInteger(ID: Integer): Integer;
    function ToBoolean(ID: Integer): Integer;
    function ToString(ID: Integer): Integer;
    function IsCallOperator(var Arg1, Arg2, Res: Integer): boolean; overload;
    function IsCallOperator: boolean; overload;
    function UndefinedID: Integer;
    procedure RemoveLastOperator;
    function LastCodeLine: Integer;
    procedure InsertCode(L1, L2: Integer);
    procedure SetVars(Vars: Integer);
//    function strIncompatibleTypes(T1, T2: Integer): String;
//    function MatchAssignment(ID1, ID2, ResID: Integer;
//                             RaiseException: Boolean = true;
//                             InitT1: Integer = 0;
//                             InitT2: Integer = 0): Integer;
//    procedure MatchTypes; virtual;
//    procedure MatchTheseTypes(S: TIntegerSet); virtual;
//    procedure CompareTypes; virtual;
    function Parse_ArgumentList(SubID: Integer; var Vars: Integer;
                                       CheckCall: Boolean = true;
                                       Erase: Boolean = true): Integer; virtual;
    function  Parse_ArrayLiteral: Integer; virtual;
    procedure Parse_ObjectInitializer(ObjectID: Integer);
    procedure GenDestroyTempObjects;
    procedure GenDestroyArrayArgumentList;
    procedure GenDestroyLocalVars; virtual;
    function  OpResultType(T1, T2: Integer): Integer;

    procedure Parse_Program; virtual; abstract;
    procedure Parse_ImportsStmt;
    function  Parse_Rank: Integer;
    function  Parse_ByRef: Integer;
    function  Parse_SetLabel: Integer;
    function  Parse_UseLabel: Integer;
    function  Parse_Constant: Integer;
    function  Parse_Ident: Integer;
    function  Parse_OverloadableOperator: Integer; virtual;
    procedure Parse_GoToStmt;
    procedure Parse_PrintList;
    procedure Parse_PrintlnList;
    function  Parse_StringLiteral: Integer;
    function  Parse_CallConv: Integer;
    procedure Parse_ReducedAssignment(LeftID: Integer);
    function  Parse_ShortEvalAND(ID: Integer;
                                Method0: TIntegerMethodNoParam;
                                Method1: TIntegerMethodOneParam): Integer;
    function Parse_ShortEvalOR(ID: Integer;
                               Method0: TIntegerMethodNoParam;
                               Method1: TIntegerMethodOneParam): Integer;

    function  BinOp(OP, Arg1, Arg2: Integer): Integer;
    function  IsOperator(OperList: TTitanIds; var OP: Integer): Boolean;
    procedure MoveUpSourceLine;
    function  IsBaseType(const S: String): Boolean; virtual;
    procedure TestDupLocalVars(NewVarID: Integer);
    function  Parse_RegExpr(const ConstructorName: String): Integer;
    function  Parse_EvalExpression: Integer; virtual; abstract;
    function  Parse_ArgumentExpression: Integer; virtual; abstract;
    procedure Parse_StmtList; virtual; abstract;

    property Name[ID: Integer]: String read GetName write SetName;
    property NameIndex[ID: Integer]: Integer read GetNameIndex write SetNameIndex;
    property Kind[ID: Integer]: Integer read GetKind write SetKind;
    property Value[ID: Integer]: Variant read GetVariant write SetVariant;
    property Address[ID: Integer]: Pointer read GetAddress write SetAddress;
    property TypeID[ID: Integer]: Integer read GetTypeID write SetTypeID;
    property Count[ID: Integer]: Integer read GetCount write SetCount;
    property Next[ID: Integer]: Integer read GetNext write SetNext;
    property TypeSub[ID: Integer]: TTitanTypeSub read GetTypeSub write SetTypeSub;
  end;

TTitanParserList = class
  private
    fList: TList;
    function GetParser(I: Integer): TTitanParser;
    function GetName(I: Integer): String;
  public
    constructor Create;
    destructor Destroy; override;
    function Count: Integer;
    function IndexOf(P: TTitanParser): Integer;
    function Add(P: TTitanParser): Integer;
    function IndexOfLanguage(const LanguageName: String): Integer;
    function GetFileExt(const LanguageName: String): String;
    function GetLanguageName(const FileName: String): String;
    function FindParser(const LanguageName: String): TTitanParser;
    procedure RemoveParser(I: Integer);
    property Parsers[I: Integer]: TTitanParser read GetParser; default;
    property Names[I: Integer]: String read GetName;
  end;

RegExp = class(TPersistent)
    private
      fLastIndex: Integer;
      fRegExpr: TRegExpr;
      scripter: Pointer;
      fZERO_BASED_STRINGS: Boolean;
      function  GetSource: Variant;
      procedure SetSource(const Value: Variant);
      function  GetInput: Variant;
      procedure SetInput(const Value: Variant);
      function  GetGlobal: Boolean;
      procedure SetGlobal(const Value: Boolean);
      function  GetIgnoreCase: Boolean;
      procedure SetIgnoreCase(const Value: Boolean);
      function  GetMultiLine: Boolean;
      procedure SetMultiLine(const Value: Boolean);
    public
      constructor Create(const Source: String = ''; const Modifiers: String = '');
      destructor Destroy; override;
      function Test(const InputString: Variant): Boolean;
      procedure Compile;
      function Exec(const InputString: Variant): TTitanArray;
      function Execute(const InputString: Variant): TTitanArray;
      function MatchCount: Integer;
      function Replace(const Expression, ReplaceStr: Variant): String;
      function toString: String;
      function RegExpr: TRegExpr;
      procedure SetScripter(s: Pointer);
    published
      property Global: Boolean read GetGlobal write SetGlobal;
      property IgnoreCase: Boolean read GetIgnoreCase write SetIgnoreCase;
      property MultiLine: Boolean read GetMultiLine write SetMultiLine;
      property LastIndex: Integer read fLastIndex write fLastIndex;
      property Source: Variant read GetSource write SetSource;
      property Input: Variant read GetInput write SetInput;
    end;

type
  TPascalScanner = class(TTitanScanner)
  public
    procedure ReadToken; override;
  end;

  TPascalParser = class
  public
    Scanner: TPascalScanner;
    CurrToken: TTitanToken;
    D: TTitanMethodDefinition;
    ResultType: String;
    NP: Integer;
    IsDynamicArrayType: Boolean;
    constructor Create;
    destructor Destroy; override;
    procedure Call_SCANNER;
    function  IsCurrText(const S: String): boolean;
    function  Parse_Ident: String;
    procedure Match(const S: String);
    function  Parse_Type(var TypeID, ExtraTypeID, Size: Integer): ShortString;
    procedure Parse_IdentList;
    procedure Parse_FormalParameter;
    procedure Parse_FormalParameters;
    procedure Parse_Heading;
    function  Parse_Property(var ReadName, WriteName: String; var Def: Boolean; DefList: TTitanDefinitionList): String;
    procedure ParseUsesClause(Output: TStrings);
    //..........................................
    procedure Parse_Heading_ForRoutine;
    procedure Parse_Heading_ForFakeMethod;
    function  Parse_PropertySXi(var Def: Boolean; DefList: TTitanDefinitionList): String;
  end;


type

TTitanJavaScriptScanner = class(TTitanScanner)
  public
    procedure ReadToken; override;
    function UsesIdent(const IdentName: String): Boolean;
  end;

TTitanJavaScriptParser = class(TTitanParser)
  private
    NewLine: Boolean;
    FunctionObjectID: Integer;
    LeftSideID: Integer;
  public
    constructor Create; override;
    destructor Destroy; override;
    procedure Reset; override;
    procedure Match(const S: String); override;
    procedure Call_SCANNER; override;
    function Gen(Op, Arg1, Arg2, Res: Integer): Integer; override;
    procedure GenDestroyLocalVars; override;

    function Parse_EvalExpression: Integer; override;
    function Parse_ArgumentExpression: Integer; override;

    function Parse_ArgumentList(SubID: Integer; var Vars: Integer;
                                CheckCall: Boolean = true;
                                Erase: Boolean = true): Integer; override;

    function Parse_PrimaryExpression: Integer;
    function Parse_ArrayLiteral: Integer; override;
    function Parse_ObjectLiteral: Integer;
    function Parse_MemberExpression(ID: Integer): Integer;
    function Parse_NewExpression: Integer;
    function Parse_Arguments: Integer;
    function Parse_LeftHandSideExpression: Integer;
    function Parse_PostfixExpression(Left: Integer): Integer;
    function Parse_UnaryExpression(Left: Integer): Integer;
    function Parse_MultiplicativeExpression(Left: Integer): Integer;
    function Parse_AdditiveExpression(Left: Integer): Integer;
    function Parse_ShiftExpression(Left: Integer): Integer;
    function Parse_RelationalExpression(Left: Integer): Integer;
    function Parse_EqualityExpression(Left: Integer): Integer;
    function Parse_BitwiseANDExpression(Left: Integer): Integer;
    function Parse_BitwiseXORExpression(Left: Integer): Integer;
    function Parse_BitwiseORExpression(Left: Integer): Integer;
    function Parse_LogicalANDExpression(Left: Integer): Integer;
    function Parse_LogicalORExpression(Left: Integer): Integer;
    function Parse_ConditionalExpression(Left: Integer): Integer;

    function Parse_AssignmentExpression: Integer;
    function Parse_Expression: Integer;

    function Parse_ModifierList: TTitanModifierList;
    procedure Parse_Statement;
    procedure Parse_Block;
    procedure Parse_IfStmt;
    procedure Parse_NamespaceStmt(ml: TTitanModifierList);
    function Parse_FunctionStmt(ml: TTitanModifierList): Integer;
    procedure Parse_ReturnStmt;
    procedure Parse_HaltStmt;
    function Parse_VarStmt(IsField: Boolean; ml: TTitanModifierList;
                           IsLoopVar: Boolean = false): Integer;
    function Parse_VariableDeclaration(IsField: Boolean; ml: TTitanModifierList;
                                       IsLoopVar: Boolean = false): Integer;
    procedure Parse_ExpressionStmt;
    procedure Parse_WhileStmt;
    procedure Parse_DoStmt;
    procedure Parse_ForStmt;
    procedure Parse_ContinueStmt;
    procedure Parse_BreakStmt;
    procedure Parse_SwitchStmt;
    procedure Parse_WithStmt;
    procedure Parse_ThrowStmt;
    procedure Parse_TryStmt;

    procedure Parse_StmtList; override;
    procedure Parse_SourceElements;

    procedure CreateGlobalObjects;
    procedure Parse_Program; override;
  end;


type
TTitanBaseScripter = class;


 TTitanModule = class(TStringList)
  public
    Name: string;
    FileName: string;
    LanguageName: string;
    Namespaces: TTitanIds;
    S1, P1, C1, S2, P2, C2: integer;
    IsSource: boolean;
    BuffStream: TStream;
    Scripter: TTitanBaseScripter;
    constructor Create(Scripter: TTitanBaseScripter);
    destructor Destroy; override;
    procedure AddNamespace(ID: integer);
    function GetTextPos(LineNumber, Position: integer): integer;
    procedure _SaveToStream(S: TStream);
    procedure _LoadFromStream(S: TStream);
  end;

 TTitanModules = class(TStringList)
  private
    Scripter: TTitanBaseScripter;
    function GetModule(Index: integer): TTitanModule;
    function GetSourceCode: string;
  public
    constructor Create(Scripter: TTitanBaseScripter);
    procedure Clear; reintroduce;
    function IndexOf(const Name: string): integer; override;
    function Add(const Name, LanguageName: string): integer; reintroduce;
    procedure Delete(Index: integer); reintroduce;
    procedure _SaveToStream(S: TStream);
    procedure _LoadFromStream(S: TStream);
    destructor Destroy; override;
    procedure Dump(const FileName: string);
    property Items[Index: integer]: TTitanModule read GetModule;
    property SourceCode: string read GetSourceCode;
  end;



  TScripterEvent = procedure(Sender: TObject) of object;
  TCompilerProgressEvent = procedure(Sender: TObject; ModuleNumber: integer) of object;
  TPrintEvent = procedure(Sender: TObject; const S: string) of object;
  TReadEvent = procedure(Sender: TObject; var S: string) of object;
  TVarEvent = procedure(Sender: TObject; ID: integer) of object;
  TVarEventEx = procedure(Sender: TObject; ID: integer; var Mode: integer) of object;
  TCodeEvent = procedure(Sender: TObject; N: integer; var Handled: boolean) of object;
  TUsedModuleEvent = procedure(Sender: TObject; const UsedModuleName, FileName: string; var SourceCode: string) of object;
  TLoadSourceCodeEvent = procedure(Sender: TObject; const UsedModuleName, FileName: string; var SourceCode: string) of object;
  TIncludeEvent = procedure(Sender: TObject; const IncludedFileName: string; var SourceCode: string) of object;
  TStreamEvent = procedure(Sender: TObject; Stream: TStream; const ModuleName: string) of object;
  TOnDefineEvent = procedure(Sender: TObject; const DirectiveName: string) of object;
  TOnChangeStateEvent = procedure(Sender: TObject; OldState, NewState: TTitanScripterState) of object;
  TOnInstanceEvent = procedure(Sender: TObject; Instance: TObject) of object;
  TLoadDllEvent = procedure(Sender: TObject; const DllName, ProcName: string; var Address: Pointer) of object;

  TVarArray = array of variant;

  TVirtualObjectMethodCallEvent = function(Sender: TObject; const ObjectName, PropName: string;
    const Params: TVarArray): variant of object;
  TVirtualObjectPutPropertyEvent = procedure(Sender: TObject; const ObjectName, PropName: string;
    const Params: TVarArray; const Value: variant) of object;

TTitanBaseScripter = class
  private
    fOnHalt: TScripterEvent;
    fOnShowError: TScripterEvent;
    fOnCompilerProgress: TCompilerProgressEvent;
    fOnRunning: TCodeEvent;
    fOnPrint: TPrintEvent;
    fOnRead: TReadEvent;
{$IFDEF UNDECLARED_EX}
    fOnUndeclaredIdentifier: TVarEventEx;
{$ELSE}
    fOnUndeclaredIdentifier: TVarEvent;
{$ENDIF}
    fOnChangedVariable: TVarEvent;
    fOnReadExtraData, fOnWriteExtraData: TStreamEvent;
    fOnUsedModule: TUsedModuleEvent;
    fOnInclude: TIncludeEvent;
    fOnLoadSourceCode: TLoadSourceCodeEvent;
    fOnDefine: TOnDefineEvent;
    fOnChangeState: TOnChangeStateEvent;
    fOnDelphiInstanceCreate: TOnInstanceEvent;
    fOnDelphiInstanceDestroy: TOnInstanceEvent;
    fOnLoadDll: TLoadDllEvent;
    fOnVirtualObjectMethodCall: TVirtualObjectMethodCallEvent;
    fOnVirtualObjectPutProperty: TVirtualObjectPutPropertyEvent;

    SignDefs: boolean;

    fOverrideHandlerMode: integer;
{$IFDEF ONRUNNING}
    // Event to replace OnRunning event.  Event will be executed when
    // fOnRunningUpdateActive is enabled and the event is assigned.
    fOnRunningUpdate: TScripterEvent;

    // Flag used to fire the OnRunningUpdate event.  This allows scripter
    // to run quickly when flag is disabled (False).  Flag can be
    // enabled/disabled during script run-time.
    fOnRunningUpdateActive: boolean;

    // Event fires when executing a function or property and UserData is non-zero.
    // Used for multi-thread applications.
    fOnRunningSync: TScripterEvent;
{$ENDIF}
    procedure SetState(Value: TTitanScripterState);
    procedure CopyLevelStack(Parser: TTitanParser);
  public
    EvalCount: integer;
    Owner: TObject;
    Modules: TTitanModules;
    SymbolTable: TTitanSymbolTable;
    Code: TTitanCode;
    ClassList: TTitanClassList;
    EventHandlerList: TTitanEventHandlerList;
    MethodBody: TTitanMethodBody;
    VariantStack: TTitanVariantStack;
    NegVarList: TTitanVarList;
    ScriptObjectList: TTitanScriptObjectList;
    ActiveXObjectList: TTitanObjectList;
    CompileTimeHeap: TTitanCompileTimeHeap;
    fState: TTitanScripterState;
    Error: variant;
    ErrorInstance: TTitanError;
    RegisteredFieldList: TTitanFieldList;
    ParserList: TTitanParserList;

    ExtraCodeList: TTitanCodeList;

    fTotalLineCount: integer;

    ParamList: TTitanParamList;

    ArrayClassRec: TTitanClassRec;

    LastResultID: integer;
    fLastResult: variant;

    DoNotDestroyList: TTitanIds;
    TempObjectList: TTitanAssocList;

    ForbiddenPublishedProperties: TList;
    ForbiddenPublishedPropertiesEx: TStringList;

    PrototypeNameIndex: integer;
    ConstructorNameIndex: integer;

    fStackSize: integer;
    fLongStrLiterals: boolean;

    CurrModule: integer;

    OwnerEventHandlerMethod: TMethod;

    AllowedEvents: boolean;
    Visited: TList;
    ExtraModuleList: TStringList;
    DefList: TStringList;

    RunList: TTitanStack;
    fSearchPathes: TStringList;
    DefaultParameterList: TDefaultParameterList;

    UnknownTypes: TTitanIDRecList;
    CallRecList: TTitanCallRecList;
    TypeAliasList: TTitanAssociativeList;
    NameList: TTitanNameList;
    LocalDefinitions: TTitanDefinitionList;

    TempVariant: variant;
    CancelMessage: string;

    _ObjCount: int64;
    LastDefinitionListCount: integer;

    IgnoreBreakpoints: boolean;

    constructor Create(Owner: TObject);
    destructor Destroy; override;
    procedure AddDefs;
    procedure AddLocalDefs;
    function AddModule(const ModuleName, LanguageName: string): integer;
    procedure AddCode(const ModuleName, Code: string);
    procedure AddCodeFromFile(const ModuleName, FileName: string);
    function CompileModule(const Name: string; Parser: TTitanParser; SyntaxCheckOnly: boolean = False): boolean;
    procedure CompileExtraCode;
    procedure Link(reallocate: boolean);
    procedure Run(RunMode: integer = _rmRun; DestLine: integer = 0);
    procedure ResetCompileStage;
    procedure ResetScripterEx;
    procedure InitRunStage;
    procedure ResetRunStage;
    procedure DiscardError;
    function IsError: boolean;
    function AddBreakpoint(const ModuleName: string; SourceLineNumber: integer; const Condition: string;
      PassCount: integer): boolean;
    function RemoveBreakpoint(const ModuleName: string; SourceLineNumber: integer; const Condition: string;
      PassCount: integer): boolean;
    procedure RemoveAllBreakpoints;
    function SourceLineToPCodeLine(const ModuleName: string; SourceLineNumber: integer): integer;
    procedure CreateErrorObject(const Message: string; PosNumber: integer);
    procedure ShowError;
    function CallMethod(SubID: integer; const This: variant; const P: array of const;
      IsEventHandler: boolean = False): variant;
    function CallMethodEx(SubID: integer; const This: variant; const P: array of const;
      const StrTypes: array of string; IsEventHandler: boolean = False): variant;
    function AssignedObject(SO: TTitanScriptObject): boolean;
    procedure DisconnectObjects;
    procedure Dump;
    function GetProperty(const ScriptObject: variant; PropertyName: string): variant;
    function Eval(const SourceCode: string; Parser: TTitanParser): variant;
    procedure EvalStatementList(const SourceCode: string; Parser: TTitanParser);
    function GetName(NameIndex: integer): string;
    procedure SaveToStream(S: TStream);
    procedure LoadFromStream(S: TStream);

    procedure SaveCompiledModuleToStream(M: TTitanModule; S: TStream);
    procedure LoadCompiledModuleFromStream(M: TTitanModule; S: TStream);

    property State: TTitanScripterState read fState write SetState;
    property TotalLineCount: integer read fTotalLineCount write fTotalLineCount;
    procedure AddExtraModule(const ModuleName, SourceCode: string; SyntaxCheckOnly: boolean;
      const PaxLanguage: string);
    procedure RunEx(ExtendedRun: boolean);
    function FindFullName(const FileName: string): string;
    function IsCompiledScript(const FileName: string): boolean;
    procedure CreateRunList;
    procedure CheckForUndeclared;
    procedure CheckCalls;
    function GetTypeID(ID: integer): integer;

    function MatchAssignment(T1, T2: integer): boolean;
    function MatchAssignmentStrict(T1, T2: integer): boolean;
    function OpResultType(T1, T2: integer): integer;
    function MatchTypes(T1, T2: integer): integer;
    function strIncompatibleTypes(T1, T2: integer): string;

    function ConvertDelphiForm1(const DfmFileName, UnitFileName, PaxLanguage: string): string;
    function ConvertDelphiForm2(const DfmFileName: string; UsedUnits: TStringList; const PaxLanguage: string): string;
    procedure AddDelphiForm(const DfmFileName, UnitFileName, PaxLanguage: string);
    function HasForbiddenPublishedProperty(C: TClass; const PropName: string): boolean;

    property OverrideHandlerMode: integer read fOverrideHandlerMode write fOverrideHandlerMode;

    property OnHalt: TScripterEvent read fOnHalt write fOnHalt;
    property OnShowError: TScripterEvent read fOnShowError write fOnShowError;
    property OnCompilerProgress: TCompilerProgressEvent read fOnCompilerProgress write fOnCompilerProgress;
    property OnPrint: TPrintEvent read fOnPrint write fOnPrint;
    property OnRead: TReadEvent read fOnRead write fOnRead;
{$IFDEF UNDECLARED_EX}
    property OnUndeclaredIdentifier: TVarEventEx read fOnUndeclaredIdentifier write fOnUndeclaredIdentifier;
{$ELSE}
    property OnUndeclaredIdentifier: TVarEvent read fOnUndeclaredIdentifier write fOnUndeclaredIdentifier;
{$ENDIF}
    property OnChangedVariable: TVarEvent read fOnChangedVariable write fOnChangedVariable;
    property OnRunning: TCodeEvent read fOnRunning write fOnRunning;
{$IFDEF ONRUNNING}
    // See variable definitions above for details.
    property OnRunningUpdate: TScripterEvent read fOnRunningUpdate write fOnRunningUpdate;
    property OnRunningUpdateActive: boolean read fOnRunningUpdateActive write fOnRunningUpdateActive;
    property OnRunningSync: TScripterEvent read fOnRunningSync write fOnRunningSync;
{$ENDIF}
    property OnReadExtraData: TStreamEvent read fOnReadExtraData write fOnReadExtraData;
    property OnWriteExtraData: TStreamEvent read fOnWriteExtraData write fOnWriteExtraData;
    property OnUsedModule: TUsedModuleEvent read fOnUsedModule write fOnUsedModule;
    property OnInclude: TIncludeEvent read fOnInclude write fOnInclude;
    property OnLoadSourceCode: TLoadSourceCodeEvent read fOnLoadSourceCode write fOnLoadSourceCode;
    property OnDefine: TOnDefineEvent read fOnDefine write fOnDefine;
    property OnChangeState: TOnChangeStateEvent read fOnChangeState write fOnChangeState;
    property OnDelphiInstanceCreate: TOnInstanceEvent read fOnDelphiInstanceCreate write fOnDelphiInstanceCreate;
    property OnDelphiInstanceDestroy: TOnInstanceEvent read fOnDelphiInstanceDestroy write fOnDelphiInstanceDestroy;
    property OnLoadDll: TLoadDllEvent read fOnLoadDll write fOnLoadDll;
    property OnVirtualObjectMethodCall: TVirtualObjectMethodCallEvent read fOnVirtualObjectMethodCall write fOnVirtualObjectMethodCall;
    property OnVirtualObjectPutProperty: TVirtualObjectPutPropertyEvent read fOnVirtualObjectPutProperty write fOnVirtualObjectPutProperty;
  end;


TCustomScripter = class(TComponent)
  protected
    fSearchPathes: TStrings;
    procedure SetSearchPathes(const Value: TStrings);
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
  published
    property SearchPathes: TStrings read fSearchPathes write SetSearchPathes;
  end;


//========================================================================

procedure Initialization_BASE_REGEXP;
procedure Initialization_BASE_EXTERN;
procedure Finalization_BASE_EXTERN;

function PaxArrayToDynamicArray(const V: Variant; ElTypeID: Integer): Pointer;
function PaxArrayToByteSet(const V: Variant): TByteSet;
function ByteSetToPaxArray(const S: TByteSet; Scripter: Pointer): Variant;
function StringToByteSet(pti: PTypeInfo; const V: String): TByteSet;
function ComparePaxArrays(A1, A2: TTitanArray): Boolean;
function UnaryPlus(const V: Variant): Variant;

function AddSets(const S1, S2: Variant): Variant;
function SubSets(const S1, S2: Variant): Variant;
function IntersectSets(const S1, S2: Variant): Variant;
function AreEqualSets(const S1, S2: Variant): Boolean;
function EqualVariants(Scripter: Pointer; const V1, V2: Variant;
                       T1: Integer = -1;
                       T2: Integer = -1): TBoolean;


procedure AddTypeAlias(const T1, T2: String);
function  FindTypeAlias(const TypeName: String; UpCase: Boolean): String;

//--------------------------------
function _GetName(NameIndex: Integer; Scripter: Pointer): String;
function _ToStr(Scripter: Pointer; const V: Variant): String;
function _Shr(X, Y: Integer): Variant;
function _Eval(const SourceCode: string; Scripter: TTitanBaseScripter; Parser: TTitanParser): variant;

function ScriptObjectToVariant(const Value: TTitanScriptObject): Variant;
function VariantToScriptObject(const Value: Variant): TTitanScriptObject;
function CreateNameIndex(const Name: String; Scripter: Pointer): Integer;
function NameIndexToUpperCaseIndex(NameIndex: Integer; Scripter: Pointer): Integer;
function DuplicateObject(const V: Variant): Variant;

function ToStr(Scripter: Pointer; const V: Variant): String;
function ToInteger(const V: Variant): Integer;
function ToObject(const Value: Variant; Scripter: Pointer): Variant;
function ToPrimitive(const V: Variant): Variant;
function ToString(const Value: Variant): String;
function ToNumber(const V: Variant): Variant;
function ToDate(const V: Variant): Variant;
function ToInt32(const V: Variant): Variant;
function ToInt64(const V: Variant): Variant;
function ToBoolean(const V: Variant): Variant;
function ToCurrency(const V: Variant): Currency;

function  RelationalComparison(const V1, V2: Variant): Variant;
function  EqualityComparison(const V1, V2: Variant): TBoolean;
function  StrictEqualityComparison(const V1, V2: Variant): TBoolean;
procedure SortVariants(var A: array of Variant);
function  StringValueToVariant(const S: String; const Dest: Variant): Variant;
function  UntypedValueToVariant(P: Pointer; Count: Integer; const Dest: Variant): Variant;
procedure VariantToUntypedValue(const V: Variant; P: Pointer; Count: Integer);
function  TVarRecToVariant(const P: TVarRec; Scripter: Pointer): Variant;


function IsNaN(const V: Variant): boolean;
function IsUndefined(const V: Variant): boolean;
function IsString(const V: Variant): boolean;
function IsNumber(const V: Variant): boolean;
function IsBoolean(const V: Variant): boolean;
function IsDateObject(const V: Variant): Boolean;
function IsHostObject(const V: Variant): Boolean;
function IsPaxArray(const V: Variant): Boolean;
function IsDynamicArray(const V: Variant): Boolean;
function InSet(const Value, ASet: Variant): Boolean;
function IsSubSet(const S1, S2: Variant): Boolean;
function IsStrictSubSet(const S1, S2: Variant): Boolean;
function IsObject(const V: Variant): boolean;
function IsStringObject(const V: Variant): Boolean;
function IsNumberObject(const V: Variant): Boolean;
function IsBooleanObject(const V: Variant): Boolean;
function IsFunctionObject(const V: Variant): Boolean;
function IsVBArray(const V: Variant): boolean;
function IsAlias(P: PVariant): Boolean;

function IsDelphiClass(Address: Pointer): Boolean;
function IsDelphiObject(Address: Pointer): Boolean;
function IsEqualBytes(P1, P2: Pointer; L: Integer): Boolean;
function IsZeroBytes(P: Pointer; L: Integer): Boolean;
function IsEqualGUID(const guid1, guid2: TGUID): Boolean;
function IsValidName(const S: String): Boolean;
function IsDigits(const S: String): Boolean;
function IsNumericString(const S: String): Boolean;

function DelphiInstanceToScriptObject(Instance: TObject; Scripter: Pointer;
                                      RaiseException: boolean = true): TTitanScriptObject;
function DelphiClassToScriptObject(AClass: TClass; Scripter: Pointer): TTitanScriptObject;
function InterfaceToScriptObject(const I: IUnknown; Scripter: Pointer;
                                 const InterfaceClassName: String = ''): TTitanScriptObject;

//--------------
procedure Call(CallAddr: Pointer;
               AClass: TClass;
               Instance: TObject;
               IsConstructor: Boolean;
               CallConv: Integer;
               Params: array of Pointer;
               Types: array of Integer;
               ExtraTypes: array of Integer;
               ByRefs: array of Boolean;
               Sizes: array of Integer;
               IsIntf: Boolean);


//------------
function StrEql(Const S1, S2: String): Boolean;
function ShiftPointer(P: Pointer; D: Integer): Pointer;
function Norm(const S: String; L: Integer): String;
function GetOperName(OP: Integer): String;
function  ArrayCreate(const P: array of Integer): Variant;
procedure ArrayPut(V: PVariant; const Indexes: array of Integer; const Val: Variant);
function  ArrayGet(V: PVariant; const Indexes: array of Integer): PVariant;
procedure ErrMessageBox(const S:  String);
procedure BackUpTextFile(const FileName: String);
function StringToBinary(const S: String): Integer;
function StringToOctal(const S: String): Integer;
function ChPos(ch: Char; const S: String): Integer;
function SetToVariantArray(Val: Integer; pti: PTypeInfo): Variant;
function HashNumber(const S: String): Integer;

procedure SetBit(var Value: Integer; const Bit: Integer);
function  TestBit(const Value: Integer; const Bit: Integer): Boolean;
procedure ClearBit(var Value: Integer; const Bit: Integer);

procedure SaveInteger(Value: Integer; S: TStream);
function  LoadInteger(S: TStream): Integer;

procedure SaveString(const Value: String; S: TStream);
function  LoadString(S: TStream): String;

procedure SaveVariant(const Value: Variant; S: TStream);
function  LoadVariant(S: TStream): Variant;

procedure AdjustEnum(var Val: Integer);

//------------------------------

function  CompareIntegers(P1, P2: Pointer): Integer;
procedure SortList(L: TList; CompareItems: TCompareItems; I1, I2: Integer);
function  GeTTitanType(const V: Variant): Integer;
function  DelphiDateTimeToEcmaTime(const AValue: TDateTime): Int64;
function  EcmaTimeToDelphiDateTime(const AValue: Variant): TDateTime;
procedure GetMethodList(FromClass: TClass; MethodList: TStrings);
procedure GetPublishedProperties(AClass: TClass; result: TStrings);
procedure GetPublishedPropertiesEx2(AClass: TClass; result: TStrings);
procedure GetPublishedEvents(AClass: TClass; result: TStrings);
procedure GetPublishedPropertiesEx(AClass: TClass; result: TStrings); // including type name
function  HasPublishedProperty(C: TClass; const PropName: String; Scripter: Pointer; NeedCheckForForbidden: Boolean = true): boolean;
function  HasPublishedPropertyEx(C: TClass; const PropName: String; var ClassType: String; Scripter: Pointer): boolean;
function  GetRefCountPtr(const S: String): Pointer;
function  GetRefCount(const S: String): Integer;
function  GetSizePtr(const S: String): Pointer;

function GetTerminal(P: PVariant): PVariant;
function CreateAlias(P: PVariant): Variant;

procedure Initialization_BASE_SYS;
procedure Finalization_BASE_SYS;

procedure SaveStringToTextFile(const S, FileName: String);
procedure ClearVar(V: PVariant);

function IntfRefCount(const I: IUnknown): Integer;

function GetImplementorOfInterface(const I: IUnknown): TObject;
function PosCh(ch: Char; const S: String): Integer;

//--------------------------

procedure ConvertDfmFile(const DfmFileName: string; UsedUnits, Output: TStrings; AsUnit: boolean = True;
                         const Src: TStrings = nil; const PaxLanguage: string = 'TitanPascal');
procedure ConvertXfmFile(const XfmFileName: string; UsedUnits, Output: TStrings; AsUnit: boolean = True;
                         const Src: TStrings = nil; const PaxLanguage: string = 'TitanPascal');
procedure SaveStr(S, FileName: string);
procedure RegisterUsedClasses;
procedure ConvDFMStringtoScript(const s: string; UsedUnits, Output: TStrings; AsUnit: boolean = True;
                         const UnitName: string = ''; const Src: TStrings = nil; const PaxLanguage: string = 'TitanPascal');

procedure ConvDFMToPaxPascalScript(const DfmFileName: string; ms: TStream; UsedUnits, Output: TStrings; AsUnit: boolean = True;
                                   Src: TStrings = nil);
procedure ConvDFMToPaxBasicScript(const DfmFileName: string; ms: TStream; UsedUnits, Output: TStrings; AsUnit: boolean = True;
                                  Src: TStrings = nil);
procedure ConvDFMToPaxCScript(const DfmFileName: string; ms: TStream; UsedUnits, Output: TStrings; AsUnit: boolean = True;
                              Src: TStrings = nil);
procedure ConvDFMToPaxJavaScriptScript(const DfmFileName: string; ms: TStream; UsedUnits, Output: TStrings;
                                       AsUnit: boolean = False; Src: TStrings = nil);

 function  GetVariantValue(Scripter, Address: Pointer; TypeID: Integer; const TypeName: String = ''): Variant;
 procedure PutVariantValue(Scripter, Address: Pointer; const AValue: Variant; typeID: Integer);
 function  DynamicArrayToVariant(Scripter, P: Pointer; const ArrayTypeName: string; ElTypeID: Integer): Variant;
 function  VariantToDynamicArray(Scripter: Pointer; const V: Variant; ElTypeID: Integer): Pointer;
 procedure EraseDynamicArray(Scripter, P: Pointer; ElTypeID: Integer);

//========================================================================

var
  __Self: TObject;
  __Scripter: TObject;
  ScripterList: TList;
  DllList: TTitanDllList;
  CurrScripter: TTitanBaseScripter;

  GlobalDefinitionList:TTitanDefinitionList;
  GlobalDefListInitialCount:LongInt=0;
  GlobalTypeAliases:Th2StringsList;
  ArrayParamMethods:Th2StringsList;
  UnresolvedTypes:Th2StringsList;
  CheckDup: Boolean = false;

  BooleanClass: TTitanScriptObjectClass = nil;
  NumberClass: TTitanScriptObjectClass = nil;
  StringClass: TTitanScriptObjectClass = nil;
  DateClass: TTitanScriptObjectClass = nil;
  FunctionClass: TTitanScriptObjectClass = nil;



implementation

//========================================================================

{$i Base_tScripter_Scanner.inc}
{$i Base_tScripter_Code.inc}
{$i Base_tScripter_Conv.inc}
{$i Base_tScripter_Class.inc}
{$i Base_tScripter_Extern.inc}
{$i Base_tScripter_Symbol.inc}
{$i Base_tScripter_Parser.inc}
{$i Base_tScripter_Event.inc}
{$i Base_tScripter_RegExp.inc}
{$i Base_tScripter_PascalParser.inc}
{$i Base_tScripter_JavaScript.inc}
{$i Base_tScripter_System.inc}
{$i Base_tScripter_DLL.inc}
{$i Base_tScripter_DMF.inc}
{$i Base_tScripter_Call.inc}
{$i Base_tScripter_Module.inc}
{$i Base_tScripter_BaseScripter.inc}
{$i Base_tScripter_CustomScripter.inc}

//========================================================================


initialization
  DllList := TTitanDllList.Create;
  ScripterList := TList.Create;

  Initialization_BASE_SYS;
  Initialization_BASE_EXTERN;
  Initialization_BASE_REGEXP;

finalization
  Finalization_BASE_EXTERN;
  Finalization_BASE_SYS;

  ScripterList.Free;
  DllList.Free;
end.
